from __future__ import annotations

import datetime
import itertools

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pac
import pytest

import daft
from daft import DataType, TimeUnit
from daft.recordbatch import MicroPartition
from daft.series import Series
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.conftest import get_tests_daft_runner_name

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())

PYTHON_TYPE_ARRAYS = {
    "int": [1, 2],
    "float": [1.0, 2.0],
    "bool": [True, False],
    "str": ["foo", "bar"],
    "binary": [b"foo", b"bar"],
    "date": [datetime.date.today(), datetime.date.today()],
    "time": [datetime.time(1, 2, 3, 4), datetime.time(5, 6, 7, 8)],
    "list": [[1, 2], [3]],
    "struct": [{"a": 1, "b": 2.0}, {"b": 3.0}],
    "empty_struct": [{}, {}],
    "nested_struct": [{"a": {"b": 1}, "c": {}}, {"a": {"b": 3}, "c": {}}],
    "null": [None, None],
    "tensor": [np.ones((2, 2), np.int64), np.ones((3, 3), np.int64)],
    # The following types are not natively supported and will be cast to Python object types.
    "timestamp": [datetime.datetime.now(), datetime.datetime.now()],
}

PYTHON_INFERRED_TYPES = {
    "int": DataType.int64(),
    "float": DataType.float64(),
    "bool": DataType.bool(),
    "str": DataType.string(),
    "binary": DataType.binary(),
    "date": DataType.date(),
    "time": DataType.time(TimeUnit.us()),
    "list": DataType.list(DataType.int64()),
    "struct": DataType.struct({"a": DataType.int64(), "b": DataType.float64()}),
    "empty_struct": DataType.struct({"": DataType.null()}),
    "nested_struct": DataType.struct(
        {"a": DataType.struct({"b": DataType.int64()}), "c": DataType.struct({"": DataType.null()})}
    ),
    "null": DataType.null(),
    "tensor": DataType.tensor(DataType.int64()),
    # The following types are not natively supported and will be cast to Python object types.
    "timestamp": DataType.timestamp(TimeUnit.us()),
}

PANDAS_INFERRED_TYPES = {
    **PYTHON_INFERRED_TYPES,
    "timestamp": DataType.timestamp(TimeUnit.ns()),
}

ROUNDTRIP_TYPES = {
    "int": pa.int64(),
    "float": pa.float64(),
    "bool": pa.bool_(),
    "str": pa.large_string(),
    "binary": pa.large_binary(),
    "date": pa.date32(),
    "time": pa.time64("us"),
    "list": pa.large_list(pa.int64()),
    "struct": pa.struct({"a": pa.int64(), "b": pa.float64()}),
    "empty_struct": pa.struct({}),
    "nested_struct": pa.struct({"a": pa.struct({"b": pa.int64()}), "c": pa.struct({})}),
    "null": pa.null(),
    "tensor": PYTHON_INFERRED_TYPES["tensor"].to_arrow_dtype(),
    # The following types are not natively supported and will be cast to Python object types.
    "timestamp": pa.timestamp("us"),
}


ARROW_TYPE_ARRAYS = {
    "int8": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.int8()),
    "int16": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.int16()),
    "int32": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.int32()),
    "int64": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.int64()),
    "uint8": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.uint8()),
    "uint16": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.uint16()),
    "uint32": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.uint32()),
    "uint64": pa.array(PYTHON_TYPE_ARRAYS["int"], pa.uint64()),
    "float32": pa.array(PYTHON_TYPE_ARRAYS["float"], pa.float32()),
    "float64": pa.array(PYTHON_TYPE_ARRAYS["float"], pa.float64()),
    "string": pa.array(PYTHON_TYPE_ARRAYS["str"], pa.string()),
    "binary": pa.array(PYTHON_TYPE_ARRAYS["binary"], pa.binary()),
    "fixed_size_binary": pa.array(PYTHON_TYPE_ARRAYS["binary"], pa.binary(3)),
    "boolean": pa.array(PYTHON_TYPE_ARRAYS["bool"], pa.bool_()),
    "date32": pa.array(PYTHON_TYPE_ARRAYS["date"], pa.date32()),
    "date64": pa.array(PYTHON_TYPE_ARRAYS["date"], pa.date64()),
    "time64_microseconds": pa.array(PYTHON_TYPE_ARRAYS["time"], pa.time64("us")),
    "time64_nanoseconds": pa.array(PYTHON_TYPE_ARRAYS["time"], pa.time64("ns")),
    "list": pa.array(PYTHON_TYPE_ARRAYS["list"], pa.list_(pa.int64())),
    "fixed_size_list": pa.array([[1, 2], [3, 4]], pa.list_(pa.int64(), 2)),
    "map": pa.array(
        [[(1.0, 1), (2.0, 2)], [(3.0, 3), (4.0, 4)]],
        pa.map_(pa.float32(), pa.int32()),
    ),
    "struct": pa.array(PYTHON_TYPE_ARRAYS["struct"], pa.struct([("a", pa.int64()), ("b", pa.float64())])),
    "empty_struct": pa.array(PYTHON_TYPE_ARRAYS["empty_struct"], pa.struct([])),
    "nested_struct": pa.array(
        PYTHON_TYPE_ARRAYS["nested_struct"],
        pa.struct(
            {
                "a": pa.struct([("b", pa.int64())]),
                "c": pa.struct([]),
            }
        ),
    ),
    "null": pa.array(PYTHON_TYPE_ARRAYS["null"], pa.null()),
    "tensor": pa.ExtensionArray.from_storage(
        ROUNDTRIP_TYPES["tensor"],
        pa.array(
            [
                {"data": PYTHON_TYPE_ARRAYS["tensor"][0].ravel(), "shape": [2, 2]},
                {"data": PYTHON_TYPE_ARRAYS["tensor"][1].ravel(), "shape": [3, 3]},
            ],
            pa.struct(
                {
                    "data": pa.large_list(pa.field("item", pa.int64())),
                    "shape": pa.large_list(pa.field("item", pa.uint64())),
                }
            ),
        ),
    ),
    # The following types are not natively supported and will be cast to Python object types.
    "timestamp": pa.array(PYTHON_TYPE_ARRAYS["timestamp"]),
}


ARROW_ROUNDTRIP_TYPES = {
    "int8": pa.int8(),
    "int16": pa.int16(),
    "int32": pa.int32(),
    "int64": pa.int64(),
    "uint8": pa.uint8(),
    "uint16": pa.uint16(),
    "uint32": pa.uint32(),
    "uint64": pa.uint64(),
    "float32": pa.float32(),
    "float64": pa.float64(),
    "string": pa.large_string(),
    "binary": pa.large_binary(),
    "fixed_size_binary": pa.binary(3),
    "boolean": pa.bool_(),
    "date32": pa.date32(),
    "date64": pa.timestamp("ms"),
    "time64_microseconds": pa.time64("us"),
    "time64_nanoseconds": pa.time64("ns"),
    "list": pa.large_list(pa.int64()),
    "fixed_size_list": pa.list_(pa.int64(), 2),
    "map": pa.map_(pa.float32(), pa.int32()),
    "struct": pa.struct([("a", pa.int64()), ("b", pa.float64())]),
    "empty_struct": pa.struct([]),
    "nested_struct": pa.struct([("a", pa.struct([("b", pa.int64())])), ("c", pa.struct([]))]),
    "null": pa.null(),
    "tensor": PYTHON_INFERRED_TYPES["tensor"].to_arrow_dtype(),
    # The following types are not natively supported and will be cast to Python object types.
    "timestamp": pa.timestamp("us"),
}

if pyarrow_supports_fixed_shape_tensor():
    arrow_tensor_dtype = pa.fixed_shape_tensor(pa.int64(), (2, 2))
    # NOTE: We don't infer fixed-shape tensors when constructing a table from Python objects, since
    # the shapes may be variable across partitions.
    # PYTHON_TYPE_ARRAYS["canonical_tensor"] = list(np.arange(8).reshape(2, 2, 2))
    # PYTHON_INFERRED_TYPES["canonical_tensor"] = DataType.tensor(DataType.int64(), (2, 2))
    # ROUNDTRIP_TYPES["canonical_tensor"] = arrow_tensor_dtype
    ARROW_ROUNDTRIP_TYPES["canonical_tensor"] = arrow_tensor_dtype
    ARROW_TYPE_ARRAYS["canonical_tensor"] = pa.FixedShapeTensorArray.from_numpy_ndarray(np.arange(8).reshape(2, 2, 2))


def _with_uuid_ext_type(uuid_ext_type) -> tuple[dict, dict]:
    if get_tests_daft_runner_name() == "ray":
        # pyarrow extension types aren't supported in Ray clusters yet.
        return ARROW_ROUNDTRIP_TYPES, ARROW_TYPE_ARRAYS
    arrow_roundtrip_types = ARROW_ROUNDTRIP_TYPES.copy()
    arrow_type_arrays = ARROW_TYPE_ARRAYS.copy()
    arrow_roundtrip_types["ext_type"] = uuid_ext_type
    storage = ARROW_TYPE_ARRAYS["binary"]
    arrow_type_arrays["ext_type"] = pa.ExtensionArray.from_storage(uuid_ext_type, storage)
    return arrow_roundtrip_types, arrow_type_arrays


def test_from_pydict_roundtrip() -> None:
    table = MicroPartition.from_pydict(PYTHON_TYPE_ARRAYS)
    assert len(table) == 2
    assert set(table.column_names()) == set(PYTHON_TYPE_ARRAYS.keys())
    for field in table.schema():
        assert field.dtype == PYTHON_INFERRED_TYPES[field.name]
    schema = pa.schema(ROUNDTRIP_TYPES)
    arrs = {}
    for col_name, col in PYTHON_TYPE_ARRAYS.items():
        if col_name == "tensor":
            arrs[col_name] = ARROW_TYPE_ARRAYS[col_name]
        else:
            arrs[col_name] = pa.array(col, type=schema.field(col_name).type)
    expected_table = pa.table(arrs, schema=schema)
    assert table.to_arrow() == expected_table


def test_from_pydict_arrow_roundtrip(uuid_ext_type) -> None:
    arrow_roundtrip_types, arrow_type_arrays = _with_uuid_ext_type(uuid_ext_type)
    table = MicroPartition.from_pydict(arrow_type_arrays)
    assert len(table) == 2
    assert set(table.column_names()) == set(arrow_type_arrays.keys())
    for field in table.schema():
        assert field.dtype == (
            DataType.from_arrow_type(arrow_type_arrays[field.name].type)
            if (field.name != "empty_struct" and field.name != "nested_struct")
            else PYTHON_INFERRED_TYPES[field.name]  # empty structs are internally represented as {"": None}
        )
    expected_table = pa.table(arrow_type_arrays).cast(pa.schema(arrow_roundtrip_types))
    assert table.to_arrow() == expected_table


def test_from_arrow_roundtrip(uuid_ext_type) -> None:
    arrow_roundtrip_types, arrow_type_arrays = _with_uuid_ext_type(uuid_ext_type)
    pa_table = pa.table(arrow_type_arrays)
    table = MicroPartition.from_arrow(pa_table)
    assert len(table) == 2
    assert set(table.column_names()) == set(arrow_type_arrays.keys())
    for field in table.schema():
        assert field.dtype == (
            DataType.from_arrow_type(arrow_type_arrays[field.name].type)
            if (field.name != "empty_struct" and field.name != "nested_struct")
            else PYTHON_INFERRED_TYPES[field.name]  # empty structs are internally represented as {"": None}
        )
    expected_table = pa.table(arrow_type_arrays).cast(pa.schema(arrow_roundtrip_types))
    assert table.to_arrow() == expected_table


def test_from_pandas_roundtrip() -> None:
    df = pd.DataFrame(PYTHON_TYPE_ARRAYS)
    table = MicroPartition.from_pandas(df)
    assert len(table) == 2
    assert set(table.column_names()) == set(PYTHON_TYPE_ARRAYS.keys())
    for field in table.schema():
        assert field.dtype == PANDAS_INFERRED_TYPES[field.name]
    # pyarrow --> pandas will insert explicit Nones within the struct fields.
    df["struct"][1]["a"] = None
    df["empty_struct"][0] = {}
    df["empty_struct"][1] = {}
    df["nested_struct"][0]["c"] = {}
    df["nested_struct"][1]["c"] = {}
    pd.testing.assert_frame_equal(table.to_pandas(), df)


def test_from_pydict_list() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": [1, 2, 3]})
    assert "a" in daft_recordbatch.column_names()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pa.array([1, 2, 3], type=pa.int64())


def test_from_pydict_np() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": np.array([1, 2, 3], dtype=np.int64)})
    assert "a" in daft_recordbatch.column_names()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pa.array([1, 2, 3], type=pa.int64())


def test_from_pydict_arrow() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": pa.array([1, 2, 3], type=pa.int8())})
    assert "a" in daft_recordbatch.column_names()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pa.array([1, 2, 3], type=pa.int8())


@pytest.mark.parametrize("list_type", [pa.list_, pa.large_list])
def test_from_pydict_arrow_list_array(list_type) -> None:
    arrow_arr = pa.array([["a", "b"], ["c"], None, [None, "d", "e"]], list_type(pa.string()))
    daft_recordbatch = MicroPartition.from_pydict({"a": arrow_arr})
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the outer list array is cast to a large list array
    # (if the outer list array wasn't already a large list in the first place), and
    # the inner string array is cast to a large string array.
    expected = arrow_arr.cast(pa.large_list(pa.large_string()))
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


def test_from_pydict_arrow_fixed_size_list_array() -> None:
    data = [["a", "b"], ["c", "d"], None, [None, "e"]]
    arrow_arr = pa.array(data, pa.list_(pa.string(), 2))
    daft_recordbatch = MicroPartition.from_pydict({"a": arrow_arr})
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the inner string array is cast to a large string array.
    expected = pa.array(data, type=pa.list_(pa.large_string(), 2))
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


def test_from_pydict_arrow_map_array() -> None:
    data = [[(1, 2.0), (3, 4.0)], None, [(5, 6.0), (7, 8.0)]]
    arrow_arr = pa.array(data, pa.map_(pa.int64(), pa.float64()))
    daft_recordbatch = MicroPartition.from_pydict({"a": arrow_arr})
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the inner string and int arrays are cast to large string and int arrays.
    expected = arrow_arr.cast(pa.map_(pa.int64(), pa.float64()))
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected
    assert daft_recordbatch.to_pydict()["a"] == data


def test_from_pydict_arrow_struct_array() -> None:
    data = [{"a": "foo", "b": "bar"}, {"b": "baz", "c": "quux"}]
    arrow_arr = pa.array(data)
    daft_recordbatch = MicroPartition.from_pydict({"a": arrow_arr})
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the inner string array is cast to a large string array.
    expected = pa.array(
        data, type=pa.struct([("a", pa.large_string()), ("b", pa.large_string()), ("c", pa.large_string())])
    )
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    reason="pyarrow extension types aren't supported on Ray clusters.",
)
def test_from_pydict_arrow_extension_array(uuid_ext_type) -> None:
    pydata = [f"{i}".encode() for i in range(6)]
    pydata[2] = None
    storage = pa.array(pydata)
    arrow_arr = pa.ExtensionArray.from_storage(uuid_ext_type, storage)
    daft_recordbatch = MicroPartition.from_pydict({"a": arrow_arr})
    assert "a" in daft_recordbatch.column_names()
    # Although Daft will internally represent the binary storage array as a large_binary array,
    # it should be cast back to the ingress extension type.
    result = daft_recordbatch.to_arrow()["a"]
    assert result.type == uuid_ext_type
    assert result.to_pylist() == arrow_arr.to_pylist()


def test_from_pydict_arrow_deeply_nested() -> None:
    # Test a struct of lists of struct of lists of strings.
    data = [{"a": [{"b": ["foo", "bar"]}]}, {"a": [{"b": ["baz", "quux"]}]}]
    arrow_arr = pa.array(data)
    daft_recordbatch = MicroPartition.from_pydict({"a": arrow_arr})
    assert "a" in daft_recordbatch.column_names()
    # Perform the expected Daft cast, where each list array is cast to a large list array and
    # the string array is cast to a large string array.
    expected = pa.array(
        data,
        type=pa.struct([("a", pa.large_list(pa.field("item", pa.struct([("b", pa.large_list(pa.large_string()))]))))]),
    )
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


@pytest.mark.parametrize(
    "data,out_dtype",
    [
        (pa.array([1, 2, None, 4], type=pa.int64()), pa.int64()),
        (pa.array(["a", "b", None, "d"], type=pa.string()), pa.large_string()),
        (pa.array([b"a", b"b", None, b"d"], type=pa.binary()), pa.large_binary()),
        (pa.array([b"a", b"b", None, b"d"], type=pa.binary(1)), pa.binary(1)),
        (pa.array([[1, 2], [3], None, [None, 4]], pa.list_(pa.int64())), pa.large_list(pa.int64())),
        (pa.array([[1, 2], [3, 4], None, [None, 6]], pa.list_(pa.int64(), 2)), pa.list_(pa.int64(), 2)),
        (
            pa.array([{"a": 1, "b": 2}, {"b": 3, "c": 4}, None, {"a": 5, "c": 6}]),
            pa.struct([("a", pa.int64()), ("b", pa.int64()), ("c", pa.int64())]),
        ),
        (
            pa.array([[(1, 2), (3, 4)], None, [(5, 6), (7, 8)]], pa.map_(pa.int64(), pa.int64())),
            pa.map_(pa.int64(), pa.int64()),
        ),
    ],
)
@pytest.mark.parametrize("chunked", [False, True])
def test_from_pydict_arrow_with_nulls_roundtrip(data, out_dtype, chunked) -> None:
    if chunked:
        data = pa.chunked_array(data)
    daft_recordbatch = MicroPartition.from_pydict({"a": data})
    assert "a" in daft_recordbatch.column_names()
    if chunked:
        data = data.combine_chunks()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pac.cast(data, out_dtype)


@pytest.mark.parametrize(
    "data,out_dtype",
    [
        # Full data.
        (pa.array([1, 2, 3, 4], type=pa.int64()), pa.int64()),
        (pa.array(["a", "b", "c", "d"], type=pa.string()), pa.large_string()),
        (pa.array([b"a", b"b", b"c", b"d"], type=pa.binary()), pa.large_binary()),
        (pa.array([b"a", b"b", b"c", b"d"], type=pa.binary(1)), pa.binary(1)),
        (pa.array([[1, 2], [3], [4, 5, 6], [None, 7]], pa.list_(pa.int64())), pa.large_list(pa.int64())),
        (pa.array([[1, 2], [3, None], [4, 5], [None, 6]], pa.list_(pa.int64(), 2)), pa.list_(pa.int64(), 2)),
        (
            pa.array([{"a": 1, "b": 2}, {"b": 3, "c": 4}, {"a": 5}, {"a": 6, "c": 7}]),
            pa.struct([("a", pa.int64()), ("b", pa.int64()), ("c", pa.int64())]),
        ),
        (
            pa.array([[(1, 2), (3, 4)], [(5, 6)], [(7, 8)]], pa.map_(pa.int64(), pa.int64())),
            pa.map_(pa.int64(), pa.int64()),
        ),
        # Contains nulls.
        (pa.array([1, 2, None, 4], type=pa.int64()), pa.int64()),
        (pa.array(["a", "b", None, "d"], type=pa.string()), pa.large_string()),
        (pa.array([b"a", b"b", None, b"d"], type=pa.binary()), pa.large_binary()),
        (pa.array([b"a", b"b", None, b"d"], type=pa.binary(1)), pa.binary(1)),
        (pa.array([[1, 2], [3], None, [None, 4]], pa.list_(pa.int64())), pa.large_list(pa.int64())),
        (pa.array([[1, 2], [3, 4], None, [None, 6]], pa.list_(pa.int64(), 2)), pa.list_(pa.int64(), 2)),
        (
            pa.array([{"a": 1, "b": 2}, {"b": 3, "c": 4}, None, {"a": 5, "c": 6}]),
            pa.struct([("a", pa.int64()), ("b", pa.int64()), ("c", pa.int64())]),
        ),
        (
            pa.array([[(1, 2), (3, 4)], None, [(7, 8)]], pa.map_(pa.int64(), pa.int64())),
            pa.map_(pa.int64(), pa.int64()),
        ),
    ],
)
@pytest.mark.parametrize("chunked", [False, True])
@pytest.mark.parametrize("slice_", list(itertools.combinations(range(4), 2)))
def test_from_pydict_arrow_sliced_roundtrip(data, out_dtype, chunked, slice_) -> None:
    offset, end = slice_
    length = end - offset
    sliced_data = data.slice(offset, length)
    if chunked:
        sliced_data = pa.chunked_array(sliced_data)
    daft_recordbatch = MicroPartition.from_pydict({"a": sliced_data})
    assert "a" in daft_recordbatch.column_names()
    if chunked:
        sliced_data = sliced_data.combine_chunks()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pac.cast(sliced_data, out_dtype)


def test_from_pydict_series() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": Series.from_arrow(pa.array([1, 2, 3], type=pa.int8()))})
    assert "a" in daft_recordbatch.column_names()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pa.array([1, 2, 3], type=pa.int8())


@pytest.mark.parametrize(
    "data,out_dtype",
    [
        # Full data.
        (pa.array([1, 2, 3, 4], type=pa.int64()), pa.int64()),
        (pa.array(["a", "b", "c", "d"], type=pa.string()), pa.large_string()),
        (pa.array([b"a", b"b", b"c", b"d"], type=pa.binary()), pa.large_binary()),
        (pa.array([b"a", b"b", b"c", b"d"], type=pa.binary(1)), pa.binary(1)),
        (pa.array([[1, 2], [3], [4, 5, 6], [None, 7]], pa.list_(pa.int64())), pa.large_list(pa.int64())),
        (pa.array([[1, 2], [3, None], [4, 5], [None, 6]], pa.list_(pa.int64(), 2)), pa.list_(pa.int64(), 2)),
        (
            pa.array([{"a": 1, "b": 2}, {"b": 3, "c": 4}, {"a": 5}, {"a": 6, "c": 7}]),
            pa.struct([("a", pa.int64()), ("b", pa.int64()), ("c", pa.int64())]),
        ),
        (
            pa.array([[(1, 2), (3, 4)], [(5, 6)], [(7, 8)]], pa.map_(pa.int64(), pa.int64())),
            pa.map_(pa.int64(), pa.int64()),
        ),
        # Contains nulls.
        (pa.array([1, 2, None, 4], type=pa.int64()), pa.int64()),
        (pa.array(["a", "b", None, "d"], type=pa.string()), pa.large_string()),
        (pa.array([b"a", b"b", None, b"d"], type=pa.binary()), pa.large_binary()),
        (pa.array([b"a", b"b", None, b"d"], type=pa.binary(1)), pa.binary(1)),
        (pa.array([[1, 2], [3], None, [None, 4]], pa.list_(pa.int64())), pa.large_list(pa.int64())),
        (pa.array([[1, 2], [3, 4], None, [None, 6]], pa.list_(pa.int64(), 2)), pa.list_(pa.int64(), 2)),
        (
            pa.array([{"a": 1, "b": 2}, {"b": 3, "c": 4}, None, {"a": 5, "c": 6}]),
            pa.struct([("a", pa.int64()), ("b", pa.int64()), ("c", pa.int64())]),
        ),
        (
            pa.array([[(1, 2), (3, 4)], None, [(7, 8)]], pa.map_(pa.int64(), pa.int64())),
            pa.map_(pa.int64(), pa.int64()),
        ),
    ],
)
@pytest.mark.parametrize("slice_", list(itertools.combinations(range(4), 2)))
def test_from_arrow_sliced_roundtrip(data, out_dtype, slice_) -> None:
    offset, end = slice_
    length = end - offset
    sliced_data = data.slice(offset, length)
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": sliced_data}))
    assert "a" in daft_recordbatch.column_names()
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == pac.cast(sliced_data, out_dtype)


@pytest.mark.parametrize("list_type", [pa.list_, pa.large_list])
def test_from_arrow_list_array(list_type) -> None:
    arrow_arr = pa.array([["a", "b"], ["c"], None, [None, "d", "e"]], list_type(pa.string()))
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": arrow_arr}))
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the outer list array is cast to a large list array
    # (if the outer list array wasn't already a large list in the first place), and
    # the inner string array is cast to a large string array.
    expected = arrow_arr.cast(pa.large_list(pa.large_string()))
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


def test_from_arrow_fixed_size_list_array() -> None:
    data = [["a", "b"], ["c", "d"], None, [None, "e"]]
    arrow_arr = pa.array(data, pa.list_(pa.string(), 2))
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": arrow_arr}))
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the inner string array is cast to a large string array.
    expected = pa.array(data, type=pa.list_(pa.large_string(), 2))
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


def test_from_arrow_struct_array() -> None:
    data = [{"a": "foo", "b": "bar"}, {"b": "baz", "c": "quux"}]
    arrow_arr = pa.array(data)
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": arrow_arr}))
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the inner string array is cast to a large string array.
    expected = pa.array(
        data, type=pa.struct([("a", pa.large_string()), ("b", pa.large_string()), ("c", pa.large_string())])
    )
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


def test_from_arrow_map_array() -> None:
    data = [[(1.0, 1), (2.0, 2)], [(3.0, 3), (4.0, 4)]]
    arrow_arr = pa.array(data, pa.map_(pa.float32(), pa.int32()))
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": arrow_arr}))
    assert "a" in daft_recordbatch.column_names()
    # Perform expected Daft cast, where the inner string and int arrays are cast to large string and int arrays.
    expected = arrow_arr.cast(pa.map_(pa.float32(), pa.int32()))
    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected
    assert daft_recordbatch.to_pydict()["a"] == data


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    reason="pyarrow extension types aren't supported on Ray clusters.",
)
def test_from_arrow_extension_array(uuid_ext_type) -> None:
    pydata = [f"{i}".encode() for i in range(6)]
    pydata[2] = None
    storage = pa.array(pydata)
    arrow_arr = pa.ExtensionArray.from_storage(uuid_ext_type, storage)
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": arrow_arr}))
    assert "a" in daft_recordbatch.column_names()
    # Although Daft will internally represent the binary storage array as a large_binary array,
    # it should be cast back to the ingress extension type.
    result = daft_recordbatch.to_arrow()["a"]
    assert result.type == uuid_ext_type
    assert result.to_pylist() == arrow_arr.to_pylist()


def test_from_arrow_deeply_nested() -> None:
    # Test a struct of lists of struct of lists of strings.
    data = [{"a": [{"b": ["foo", "bar"]}]}, {"a": [{"b": ["baz", "quux"]}]}]
    arrow_arr = pa.array(data)
    daft_recordbatch = MicroPartition.from_arrow(pa.table({"a": arrow_arr}))
    assert "a" in daft_recordbatch.column_names()
    # Perform the expected Daft cast, where each list array is cast to a large list array and
    # the string array is cast to a large string array.
    expected = pa.array(
        data,
        type=pa.struct(
            [
                (
                    "a",
                    pa.large_list(
                        pa.field("item", pa.struct([("b", pa.large_list(pa.field("item", pa.large_string())))]))
                    ),
                )
            ]
        ),
    )

    assert daft_recordbatch.to_arrow()["a"].combine_chunks() == expected


def test_from_pydict_bad_input() -> None:
    with pytest.raises(ValueError, match="Mismatch in Series lengths"):
        MicroPartition.from_pydict({"a": [1, 2, 3, 4], "b": [5, 6, 7]})


def test_pyobjects_roundtrip() -> None:
    o0, o1 = object(), object()
    table = MicroPartition.from_pydict({"objs": [o0, o1, None]})
    objs = table.to_pydict()["objs"]
    assert objs[0] is o0
    assert objs[1] is o1
    assert objs[2] is None


@pytest.mark.parametrize("levels", list(range(6)))
def test_nested_list_dates(levels: int) -> None:
    data = [datetime.date.today(), datetime.date.today()]
    for _ in range(levels):
        data = [data, data]
    table = MicroPartition.from_pydict({"item": data})
    back_again = table.get_column("item")

    dtype = back_again.datatype()

    expected_dtype = DataType.date()
    expected_arrow_type = pa.date32()
    for _ in range(levels):
        expected_dtype = DataType.list(expected_dtype)
        expected_arrow_type = pa.large_list(pa.field("item", expected_arrow_type))

    assert dtype == expected_dtype

    assert back_again.to_arrow().type == expected_arrow_type
    assert back_again.to_pylist() == data


# TODO: Fix downcasting for FixedSizedList in Series::try_from
@pytest.mark.skip()
@pytest.mark.parametrize("levels", list(range(1, 6)))
def test_nested_fixed_size_list_dates(levels: int) -> None:
    data = [datetime.date.today(), datetime.date.today()]
    for _ in range(levels):
        data = [data, data]

    expected_dtype = DataType.date()
    expected_arrow_type = pa.date32()
    for _ in range(levels):
        expected_dtype = DataType.fixed_size_list(expected_dtype, 2)
        expected_arrow_type = pa.list_(expected_arrow_type, 2)

    pa_data = pa.array(data, type=expected_arrow_type)
    table = MicroPartition.from_pydict({"data": pa_data})
    back_again = table.get_column("data")

    dtype = back_again.datatype()

    assert dtype == expected_dtype
    assert back_again.to_arrow().type == expected_arrow_type
    assert back_again.to_pylist() == data


@pytest.mark.parametrize("levels", list(range(5)))
def test_nested_struct_dates(levels: int) -> None:
    data = datetime.date.today()
    for _ in range(levels):
        data = {"data": data}

    data = [data]
    table = MicroPartition.from_pydict({"data": data})
    back_again = table.get_column("data")
    dtype = back_again.datatype()
    expected_dtype = DataType.date()
    expected_arrow_type = pa.date32()
    for _ in range(levels):
        expected_dtype = DataType.struct({"data": expected_dtype})
        expected_arrow_type = pa.struct([("data", expected_arrow_type)])
    assert dtype == expected_dtype

    assert back_again.to_arrow().type == expected_arrow_type
    assert back_again.to_pylist() == data


def test_from_arrow_iterable() -> None:
    class CustomIterable:
        def __iter__(self):
            yield pa.table({"text": ["foo1", "bar2"]})
            yield pa.table({"text": ["foo2", "bar2"]})
            yield pa.table({"text": ["foo3", "bar3"]})

    my_iter = CustomIterable()

    table = daft.from_arrow(my_iter)
    tbl = table.to_pydict()
    assert tbl == {"text": ["foo1", "bar2", "foo2", "bar2", "foo3", "bar3"]}


def test_to_arrow_iterator() -> None:
    df = daft.from_pydict({"a": [1, 2, 3], "b": [4, 5, 6]})
    it = df.to_arrow_iter()
    assert isinstance(next(it), pa.RecordBatch)



import pytest

import daft
from daft import col


@pytest.mark.parametrize("num_hashes", [1, 2, 16, 128])
@pytest.mark.parametrize("ngram_size", [1, 2, 4, 5, 100])
@pytest.mark.parametrize("seed", [1, -1, 123, None])
@pytest.mark.parametrize("hash_function", ["murmurhash3", "xxhash", "sha1"])
def test_table_expr_minhash(num_hashes, ngram_size, seed, hash_function):
    df = daft.from_pydict(
        {
            "data": [
                "The quick brown fox",
                "The speedy orange fox",
                "The quick brown fox",
                "thisonlyhasonetokenohno",
                None,
                "This has more than four tokens, unlike the other strings",
                "!@# $%^&*() -` 1235 827 9387 216340",
                None,
            ]
        }
    )

    res = None
    if seed is None:
        res = df.select(col("data").minhash(num_hashes, ngram_size, hash_function=hash_function))
    else:
        res = df.select(col("data").minhash(num_hashes, ngram_size, seed, hash_function=hash_function))
    minhash = res.to_pydict()["data"]
    assert minhash[4] is None and minhash[-1] is None
    for lst in minhash:
        if lst is not None:
            assert len(lst) == num_hashes
    for i in range(num_hashes):
        assert minhash[0][i] == minhash[2][i]
        if ngram_size > 1:
            assert minhash[0][i] != minhash[1][i]



from daft.catalog import Catalog


def test_try_from_iceberg(tmpdir):
    from pyiceberg.catalog.sql import SqlCatalog

    pyiceberg_catalog = SqlCatalog(
        "default",
        **{
            "uri": f"sqlite:///{tmpdir}/pyiceberg_catalog.db",
            "warehouse": f"file://{tmpdir}",
        },
    )
    # assert doesn't throw!
    assert Catalog._from_obj(pyiceberg_catalog) is not None
    assert Catalog.from_iceberg(pyiceberg_catalog) is not None


def test_try_from_unity():
    from daft.unity_catalog import UnityCatalog

    unity_catalog = UnityCatalog("-", token="-")
    # assert doesn't throw!
    assert Catalog._from_obj(unity_catalog) is not None
    assert Catalog.from_unity(unity_catalog) is not None


def test_register_python_catalog():
    import daft.unity_catalog

    # sanity check for backwards compatibility
    cat1 = daft.unity_catalog.UnityCatalog("", "")
    daft.catalog.register_python_catalog(cat1, "test")
    daft.catalog.unregister_catalog("test")



from __future__ import annotations

import datetime

import pytest

from daft.io.object_store_options import io_config_to_storage_options

deltalake = pytest.importorskip("deltalake")

import pyarrow as pa
import pyarrow.compute as pc
import pytest

import daft
from daft.logical.schema import Schema
from tests.utils import assert_pyarrow_tables_equal

PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(
    PYARROW_LOWER_BOUND_SKIP,
    reason="deltalake not supported on older versions of pyarrow",
)


def test_read_predicate_pushdown_on_data(deltalake_table):
    deltalake = pytest.importorskip("deltalake")
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    df = df.where(df["a"] == 2)
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables([table.filter(pc.field("a") == 2) for table in tables]).sort_by("part_idx"),
    )


def test_read_predicate_pushdown_on_part(deltalake_table, partition_generator):
    deltalake = pytest.importorskip("deltalake")
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    part_idx = 2
    partition_generator, _ = partition_generator
    part_value = partition_generator(part_idx)
    if part_value is None:
        part_value = part_idx
    df = df.where(df["part_idx"] == part_value)
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables([table.filter(pc.field("part_idx") == part_value) for table in tables]),
    )


def test_read_predicate_pushdown_on_part_non_eq(deltalake_table, partition_generator):
    deltalake = pytest.importorskip("deltalake")
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    part_idx = 3
    partition_generator, _ = partition_generator
    part_value = partition_generator(part_idx)
    if part_value is None:
        part_value = part_idx
    df = df.where(df["part_idx"] < part_value)
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables([table.filter(pc.field("part_idx") < part_value) for table in tables]),
    )


def test_read_predicate_pushdown_on_part_and_data(deltalake_table, partition_generator):
    deltalake = pytest.importorskip("deltalake")
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    part_idx = 2
    partition_generator, _ = partition_generator
    part_value = partition_generator(part_idx)
    if part_value is None:
        part_value = part_idx
    df = df.where((df["part_idx"] == part_value) & (df["f"] == datetime.datetime(2024, 2, 11)))
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables(
            [
                table.filter((pc.field("part_idx") == part_value) & (pc.field("f") == datetime.datetime(2024, 2, 11)))
                for table in tables
            ]
        ),
    )


def test_read_predicate_pushdown_on_part_and_data_same_clause(deltalake_table, partition_generator):
    deltalake = pytest.importorskip("deltalake")
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    partition_generator, col = partition_generator
    df = df.where(df["part_idx"] < df[col])
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables([table.filter(pc.field("part_idx") < pc.field(col)) for table in tables]),
    )


def test_read_predicate_pushdown_on_part_empty(deltalake_table, partition_generator, num_partitions):
    deltalake = pytest.importorskip("deltalake")
    partition_generator, _ = partition_generator
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)
    # There should only be num_partitions partitions; see local_deltalake_table fixture.
    part_value = partition_generator(num_partitions)
    if part_value is None:
        part_value = num_partitions
    df = df.where(df["part_idx"] == part_value)
    delta_schema = deltalake.DeltaTable(path, storage_options=io_config_to_storage_options(io_config, path)).schema()
    expected_schema = Schema.from_pyarrow_schema(delta_schema.to_pyarrow())
    assert df.schema() == expected_schema
    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables([table.filter(pc.field("part_idx") == part_value) for table in tables]),
    )


def test_read_select_partition_key(deltalake_table):
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)

    df = df.select("part_idx", "a")

    assert df.schema().column_names() == ["part_idx", "a"]

    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by([("part_idx", "ascending"), ("a", "ascending")]),
        pa.concat_tables([table.select(["part_idx", "a"]) for table in tables]).sort_by(
            [("part_idx", "ascending"), ("a", "ascending")]
        ),
    )


def test_read_select_partition_key_with_filter(deltalake_table):
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)

    df = df.select("part_idx", "a")
    df = df.where(df["a"] < 5)

    assert df.schema().column_names() == ["part_idx", "a"]

    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by([("part_idx", "ascending"), ("a", "ascending")]),
        pa.concat_tables([table.select(["part_idx", "a"]) for table in tables]).sort_by(
            [("part_idx", "ascending"), ("a", "ascending")]
        ),
    )


@pytest.mark.skip(
    reason="Selecting just the partition key in a deltalake table is not yet supported. "
    "Issue: https://github.com/Eventual-Inc/Daft/issues/2129"
)
def test_read_select_only_partition_key(deltalake_table):
    path, catalog_table, io_config, tables = deltalake_table
    df = daft.read_deltalake(str(path) if catalog_table is None else catalog_table, io_config=io_config)

    df = df.select("part_idx")

    assert df.schema().column_names() == ["part_idx"]

    assert_pyarrow_tables_equal(
        df.to_arrow().sort_by("part_idx"),
        pa.concat_tables([table.select(["part_idx"]) for table in tables]).sort_by("part_idx"),
    )



import datetime

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_utf8_to_datetime():
    table = MicroPartition.from_pydict({"col": ["2021-01-01 00:00:00", None, "2021-01-02 00:00:00"]})
    result = table.eval_expression_list([col("col").str.to_datetime("%Y-%m-%d %H:%M:%S")])
    assert result.to_pydict() == {
        "col": [
            datetime.datetime(2021, 1, 1, 0, 0),
            None,
            datetime.datetime(2021, 1, 2, 0, 0),
        ]
    }



import pytest

import daft
from daft.sql import SQLCatalog


def test_sql_catalog_sanity():
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    catalog = {"a.b": df1}
    try:
        actual = daft.sql('select * from "a.b"', catalog=SQLCatalog(catalog)).to_pydict()
        assert actual == df1.to_pydict()
    except Exception as e:
        assert False, f"Unexpected exception: {e}"


@pytest.mark.parametrize(
    "table_name",
    ["a", "a_b", "a.b", "a.b.c", "a_", "a_1", "a-b", "a."],
)
def test_sql_catalog_table_names(table_name):
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    catalog = {table_name: df1}
    try:
        actual = daft.sql(f'select * from "{table_name}"', catalog=SQLCatalog(catalog)).to_pydict()
        assert actual == df1.to_pydict()
    except Exception as e:
        assert False, f"Unexpected exception: {e}"


@pytest.mark.parametrize(
    "table_name",
    [
        "1",  # invalid start
        "a.",  # invalid part <period>
        "a b",  # invalid part <space>
        "a-b",  # invalid part -
    ],
)
def test_sql_catalog_table_names_invalid(table_name):
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    catalog = SQLCatalog({table_name: df1})
    # ok when delimited
    _ = daft.sql(f'select * from "{table_name}"', catalog).to_pydict()
    # err when not delimited
    with pytest.raises(Exception):
        _ = daft.sql(f"select * from {table_name}", catalog).to_pydict()


def test_sql_register_globals():
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    catalog = SQLCatalog({"df2": df1})
    try:
        daft.sql("select * from df1", catalog=catalog, register_globals=False).collect()
    except Exception as e:
        assert True, f"Expected exception: {e}"



from __future__ import annotations

import pytest

pyiceberg = pytest.importorskip("pyiceberg")
import itertools
from datetime import date, datetime

import pandas as pd
import pytz

import daft
from daft.expressions import Expression
from tests.conftest import assert_df_equals


@pytest.mark.integration()
def test_daft_iceberg_table_predicate_pushdown_days(local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_days")
    df = daft.read_table(f"{catalog_name}.default.test_partitioned_by_days")
    df = df.where(df["ts"] < date(2023, 3, 6))
    df.collect()
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    # need to use datetime here
    iceberg_pandas = iceberg_pandas[iceberg_pandas["ts"] < datetime(2023, 3, 6, tzinfo=pytz.utc)]

    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


def is_null(obj):
    if isinstance(obj, Expression):
        return obj.is_null()
    elif isinstance(obj, pd.Series):
        return obj.isnull()
    else:
        raise NotImplementedError()


def udf_func(obj):
    if isinstance(obj, Expression):
        return obj.apply(lambda x: str(x)[:1] == "1", return_dtype=daft.DataType.bool())
    elif isinstance(obj, pd.Series):
        return obj.apply(lambda x: str(x)[:1] == "1")
    else:
        raise NotImplementedError()


@pytest.mark.integration()
@pytest.mark.parametrize(
    "predicate, table, limit",
    itertools.product(
        [
            lambda x: x < date(2023, 3, 6),
            lambda x: x == date(2023, 3, 6),
            lambda x: x > date(2023, 3, 6),
            lambda x: x != date(2023, 3, 6),
            lambda x: date(2023, 3, 6) > x,
            lambda x: date(2023, 3, 6) == x,
            lambda x: date(2023, 3, 6) < x,
            lambda x: date(2023, 3, 6) != x,
            is_null,
            udf_func,
        ],
        [
            "test_partitioned_by_months",
            "test_partitioned_by_years",
        ],
        [None, 1, 2, 1000],
    ),
)
def test_daft_iceberg_table_predicate_pushdown_on_date_column(predicate, table, limit, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table}")
    df = daft.read_table(f"{catalog_name}.default.{table}")
    df = df.where(predicate(df["dt"]))
    if limit:
        df = df.limit(limit)
    df.collect()

    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[predicate(iceberg_pandas["dt"])]
    if limit:
        iceberg_pandas = iceberg_pandas[:limit]
    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
@pytest.mark.parametrize(
    "predicate, table, limit",
    itertools.product(
        [
            lambda x: x < datetime(2023, 3, 6, tzinfo=pytz.utc),
            lambda x: x == datetime(2023, 3, 6, tzinfo=pytz.utc),
            lambda x: x > datetime(2023, 3, 6, tzinfo=pytz.utc),
            lambda x: x != datetime(2023, 3, 6, tzinfo=pytz.utc),
            lambda x: datetime(2023, 3, 6, tzinfo=pytz.utc) > x,
            lambda x: datetime(2023, 3, 6, tzinfo=pytz.utc) == x,
            lambda x: datetime(2023, 3, 6, tzinfo=pytz.utc) < x,
            lambda x: datetime(2023, 3, 6, tzinfo=pytz.utc) != x,
            is_null,
            udf_func,
        ],
        [
            "test_partitioned_by_days",
            "test_partitioned_by_hours",
            "test_partitioned_by_identity",
        ],
        [None, 1, 2, 1000],
    ),
)
def test_daft_iceberg_table_predicate_pushdown_on_timestamp_column(predicate, table, limit, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table}")
    df = daft.read_table(f"{catalog_name}.default.{table}")
    df = df.where(predicate(df["ts"]))
    if limit:
        df = df.limit(limit)
    df.collect()

    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[predicate(iceberg_pandas["ts"])]
    if limit:
        iceberg_pandas = iceberg_pandas[:limit]

    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
@pytest.mark.parametrize(
    "predicate, table, limit",
    itertools.product(
        [
            lambda x: x < "d",
            lambda x: x == "d",
            lambda x: x > "d",
            lambda x: x != "d",
            lambda x: x == "z",
            lambda x: "d" > x,
            lambda x: "d" == x,
            lambda x: "d" < x,
            lambda x: "d" != x,
            lambda x: "z" == x,
            is_null,
            udf_func,
        ],
        [
            "test_partitioned_by_truncate",
        ],
        [None, 1, 2, 1000],
    ),
)
def test_daft_iceberg_table_predicate_pushdown_on_letter(predicate, table, limit, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table}")
    df = daft.read_table(f"{catalog_name}.default.{table}")
    df = df.where(predicate(df["letter"]))
    if limit:
        df = df.limit(limit)
    df.collect()
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[predicate(iceberg_pandas["letter"])]
    if limit:
        iceberg_pandas = iceberg_pandas[:limit]

    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
@pytest.mark.parametrize(
    "predicate, table, limit",
    itertools.product(
        [
            lambda x: x < 4,
            lambda x: x == 4,
            lambda x: x > 4,
            lambda x: x != 4,
            lambda x: x == 100,
            lambda x: 4 > x,
            lambda x: 4 == x,
            lambda x: 4 < x,
            lambda x: 4 != x,
            lambda x: 100 == x,
            is_null,
            udf_func,
        ],
        [
            "test_partitioned_by_bucket",
        ],
        [None, 1, 2, 1000],
    ),
)
def test_daft_iceberg_table_predicate_pushdown_on_number(predicate, table, limit, local_iceberg_catalog):
    catalog_name, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table(f"default.{table}")
    df = daft.read_table(f"{catalog_name}.default.{table}")
    df = df.where(predicate(df["number"]))
    if limit:
        df = df.limit(limit)
    df.collect()
    daft_pandas = df.to_pandas()
    iceberg_pandas = tab.scan().to_arrow().to_pandas()
    iceberg_pandas = iceberg_pandas[predicate(iceberg_pandas["number"])]
    if limit:
        iceberg_pandas = iceberg_pandas[:limit]

    assert_df_equals(daft_pandas, iceberg_pandas, sort_key=[])


@pytest.mark.integration()
def test_daft_iceberg_table_predicate_pushdown_empty_scan(local_iceberg_catalog):
    _, pyiceberg_catalog = local_iceberg_catalog
    tab = pyiceberg_catalog.load_table("default.test_partitioned_by_months")
    df = daft.read_iceberg(tab)
    df = df.where(df["dt"] > date(2030, 1, 1))
    df.collect()
    values = df.to_arrow()
    assert len(values) == 0



import datetime

import daft
from daft.sql.sql import SQLCatalog


def test_temporals():
    df = daft.from_pydict(
        {
            "datetimes": [
                datetime.datetime(2021, 1, 1, 23, 59, 58),
                datetime.datetime(2021, 1, 2, 0, 0, 0),
                datetime.datetime(2021, 1, 2, 1, 2, 3),
                datetime.datetime(2021, 1, 2, 1, 2, 3),
                datetime.datetime(1999, 1, 1, 1, 1, 1),
                None,
            ]
        }
    )
    catalog = SQLCatalog({"test": df})

    expected = df.select(
        daft.col("datetimes").dt.date().alias("date"),
        daft.col("datetimes").dt.day().alias("day"),
        daft.col("datetimes").dt.day_of_week().alias("day_of_week"),
        daft.col("datetimes").dt.hour().alias("hour"),
        daft.col("datetimes").dt.minute().alias("minute"),
        daft.col("datetimes").dt.month().alias("month"),
        daft.col("datetimes").dt.second().alias("second"),
        daft.col("datetimes").dt.year().alias("year"),
    ).collect()

    actual = daft.sql(
        """
    SELECT
        date(datetimes) as date,
        day(datetimes) as day,
        dayofweek(datetimes) as day_of_week,
        hour(datetimes) as hour,
        minute(datetimes) as minute,
        month(datetimes) as month,
        second(datetimes) as second,
        year(datetimes) as year,
    FROM test
    """,
        catalog=catalog,
    ).collect()

    assert actual.to_pydict() == expected.to_pydict()


def test_extract():
    df = daft.from_pydict(
        {
            "datetimes": [
                datetime.datetime(2021, 1, 1, 23, 59, 58),
                datetime.datetime(2021, 1, 2, 0, 0, 0),
                datetime.datetime(2021, 1, 2, 1, 2, 3),
                datetime.datetime(2021, 1, 2, 1, 2, 3),
                datetime.datetime(1999, 1, 1, 1, 1, 1),
                None,
            ]
        }
    )

    expected = df.select(
        daft.col("datetimes").dt.date().alias("date"),
        daft.col("datetimes").dt.day().alias("day"),
        daft.col("datetimes").dt.day_of_week().alias("day_of_week"),
        daft.col("datetimes").dt.hour().alias("hour"),
        daft.col("datetimes").dt.minute().alias("minute"),
        daft.col("datetimes").dt.month().alias("month"),
        daft.col("datetimes").dt.second().alias("second"),
        daft.col("datetimes").dt.year().alias("year"),
    ).collect()

    actual = daft.sql("""
    SELECT
        extract(date from datetimes) as date,
        extract(day from datetimes) as day,
        extract(dayofweek from datetimes) as day_of_week,
        extract(hour from datetimes) as hour,
        extract(minute from datetimes) as minute,
        extract(month from datetimes) as month,
        extract(second from datetimes) as second,
        extract(year from datetimes) as year,
    FROM df
    """).collect()

    assert actual.to_pydict() == expected.to_pydict()


def test_date_comparison():
    date_df = daft.from_pydict({"date_str": ["2020-01-01", "2020-01-02", "2020-01-03"]})
    date_df = date_df.with_column("date", daft.col("date_str").str.to_date("%Y-%m-%d"))
    expected = date_df.filter(daft.col("date") == "2020-01-01").select("date").to_pydict()
    actual = daft.sql("select date from date_df where date == '2020-01-01'").to_pydict()
    assert actual == expected



from packaging.utils import parse_wheel_filename


def get_platform_tag(wheelname: str) -> str:
    _, _, _, tags = parse_wheel_filename(wheelname)
    assert len(tags) == 1, "Multiple tags found"
    return next(iter(tags)).platform



from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from daft.daft import PartitionField, PartitionTransform, Pushdowns, ScanTask

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.logical.schema import Field, Schema


def make_partition_field(
    field: Field, source_field: Field | None = None, transform: PartitionTransform | None = None
) -> PartitionField:
    return PartitionField(
        field._field,
        source_field._field if source_field is not None else None,
        transform,
    )


class ScanOperator(abc.ABC):
    @abc.abstractmethod
    def schema(self) -> Schema:
        raise NotImplementedError()

    @abc.abstractmethod
    def display_name(self) -> str:
        return self.__class__.__name__

    @abc.abstractmethod
    def partitioning_keys(self) -> list[PartitionField]:
        raise NotImplementedError()

    @abc.abstractmethod
    def can_absorb_filter(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def can_absorb_limit(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def can_absorb_select(self) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def multiline_display(self) -> list[str]:
        raise NotImplementedError()

    @abc.abstractmethod
    def to_scan_tasks(self, pushdowns: Pushdowns) -> Iterator[ScanTask]:
        raise NotImplementedError()



from __future__ import annotations

import dataclasses
import warnings
from typing import Callable, Literal
from urllib.parse import urlparse

import unitycatalog

from daft.io import AzureConfig, IOConfig, S3Config


@dataclasses.dataclass(frozen=True)
class UnityCatalogTable:
    table_uri: str
    io_config: IOConfig | None


class UnityCatalog:
    """Client to access the Unity Catalog.

    Unity Catalog is an open-sourced data catalog that can be self-hosted, or hosted by Databricks.

    Example of reading a dataframe from a table in Unity Catalog hosted by Databricks:

    >>> cat = UnityCatalog("https://<databricks_workspace_id>.cloud.databricks.com", token="my-token")
    >>> table = cat.load_table("my_catalog.my_schema.my_table")
    >>> df = daft.read_deltalake(table)
    """

    def __init__(self, endpoint: str, token: str | None = None):
        self._client = unitycatalog.Unitycatalog(
            base_url=endpoint.rstrip("/") + "/api/2.1/unity-catalog/",
            default_headers={"Authorization": f"Bearer {token}"},
        )

    def _paginate_to_completion(
        self,
        client_func_call: Callable[[unitycatalog.Unitycatalog, str | None], tuple[list[str] | None, str | None]],
    ) -> list[str]:
        results = []

        # Make first request
        new_results, next_page_token = client_func_call(self._client, None)
        if new_results is not None:
            results.extend(new_results)

        # Exhaust pages
        while next_page_token is not None and next_page_token != "":
            new_results, next_page_token = client_func_call(self._client, next_page_token)
            if new_results is not None:
                results.extend(new_results)

        return results

    def list_catalogs(self) -> list[str]:
        def _paginated_list_catalogs(client: unitycatalog.Unitycatalog, page_token: str | None):
            response = client.catalogs.list(page_token=page_token)
            next_page_token = response.next_page_token
            if response.catalogs is None:
                return None, next_page_token
            return [c.name for c in response.catalogs], next_page_token

        return self._paginate_to_completion(_paginated_list_catalogs)

    def list_schemas(self, catalog_name: str) -> list[str]:
        def _paginated_list_schemas(client: unitycatalog.Unitycatalog, page_token: str | None):
            response = client.schemas.list(catalog_name=catalog_name, page_token=page_token)
            next_page_token = response.next_page_token
            if response.schemas is None:
                return None, next_page_token
            return [s.full_name for s in response.schemas], next_page_token

        return self._paginate_to_completion(_paginated_list_schemas)

    def list_tables(self, schema_name: str):
        if schema_name.count(".") != 1:
            raise ValueError(
                f"Expected fully-qualified schema name with format `catalog_name`.`schema_name`, but received: {schema_name}"
            )

        catalog_name, schema_name = schema_name.split(".")

        def _paginated_list_tables(client: unitycatalog.Unitycatalog, page_token: str | None):
            response = client.tables.list(catalog_name=catalog_name, schema_name=schema_name, page_token=page_token)
            next_page_token = response.next_page_token
            if response.tables is None:
                return None, next_page_token
            return [f"{t.catalog_name}.{t.schema_name}.{t.name}" for t in response.tables], next_page_token

        return self._paginate_to_completion(_paginated_list_tables)

    def load_table(
        self,
        table_name: str,
        new_table_storage_path: str | None = None,
        operation: Literal["READ" | "READ_WRITE"] = "READ_WRITE",
    ) -> UnityCatalogTable:
        """Loads an existing Unity Catalog table. If the table is not found, and information is provided in the method to create a new table, a new table will be attempted to be registered.

        Args:
            table_name (str): Name of the table in Unity Catalog in the form of dot-separated, 3-level namespace
            new_table_storage_path (str, optional): Cloud storage path URI to register a new external table using this path. Unity Catalog will validate if the path is valid and authorized for the principal, else will raise an exception.
            operation ("READ" or "READ_WRITE", optional): The intended use of the table, which impacts authorization. Defaults to "READ_WRITE".

        Returns:
            UnityCatalogTable
        """
        # Load the table ID
        try:
            table_info = self._client.tables.retrieve(table_name)
            if new_table_storage_path:
                warnings.warn(
                    f"Table {table_name} is an existing storage table with a valid storage path. The 'new_table_storage_path' argument provided will be ignored."
                )
        except unitycatalog.NotFoundError:
            if not new_table_storage_path:
                raise ValueError(
                    f"Table {table_name} is not an existing table. If a new table needs to be created, provide 'new_table_storage_path' value."
                )
            try:
                three_part_namesplit = table_name.split(".")
                if len(three_part_namesplit) != 3 or not all(three_part_namesplit):
                    raise ValueError(
                        f"Expected table name to be in the format of 'catalog.schema.table', received: {table_name}"
                    )

                params = {
                    "catalog_name": three_part_namesplit[0],
                    "schema_name": three_part_namesplit[1],
                    "name": three_part_namesplit[2],
                    "columns": None,
                    "data_source_format": "DELTA",
                    "table_type": "EXTERNAL",
                    "storage_location": new_table_storage_path,
                    "comment": None,
                }

                table_info = self._client.tables.create(**params)
            except Exception as e:
                raise Exception(f"An error occurred while registering the table in Unity Catalog: {e}")

        table_id = table_info.table_id
        storage_location = table_info.storage_location
        # Grab credentials from Unity catalog and place it into the Table
        temp_table_credentials = self._client.temporary_table_credentials.create(operation=operation, table_id=table_id)

        scheme = urlparse(storage_location).scheme
        if scheme == "s3" or scheme == "s3a":
            aws_temp_credentials = temp_table_credentials.aws_temp_credentials
            io_config = (
                IOConfig(
                    s3=S3Config(
                        key_id=aws_temp_credentials.access_key_id,
                        access_key=aws_temp_credentials.secret_access_key,
                        session_token=aws_temp_credentials.session_token,
                    )
                )
                if aws_temp_credentials is not None
                else None
            )
        elif scheme == "gcs" or scheme == "gs":
            # TO-DO: gather GCS credential vending assets from Unity and construct 'io_config``
            warnings.warn("GCS credential vending from Unity Catalog is not yet supported.")
            io_config = None
        elif scheme == "az" or scheme == "abfs" or scheme == "abfss":
            io_config = IOConfig(
                azure=AzureConfig(sas_token=temp_table_credentials.azure_user_delegation_sas.get("sas_token"))
            )
        else:
            warnings.warn(f"Credentials for scheme {scheme} are not yet supported.")
            io_config = None

        return UnityCatalogTable(
            table_uri=storage_location,
            io_config=io_config,
        )



from __future__ import annotations

import pytest

from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition

table = MicroPartition.from_pydict({"a": [[1, 2], [3, 4], [5, None], [None, None], None]})
fixed_dtype = DataType.fixed_size_list(DataType.int64(), 2)
fixed_table = table.eval_expression_list([col("a").cast(fixed_dtype)])


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_sum(table):
    result = table.eval_expression_list([col("a").list.sum()])
    assert result.to_pydict() == {"a": [3, 7, 5, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_mean(table):
    result = table.eval_expression_list([col("a").list.mean()])
    assert result.to_pydict() == {"a": [1.5, 3.5, 5, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_min(table):
    result = table.eval_expression_list([col("a").list.min()])
    assert result.to_pydict() == {"a": [1, 3, 5, None, None]}


@pytest.mark.parametrize("table", [table, fixed_table])
def test_list_max(table):
    result = table.eval_expression_list([col("a").list.max()])
    assert result.to_pydict() == {"a": [2, 4, 5, None, None]}



# /// script
# requires-python = ">=3.12"
# dependencies = ["ray[default]"]
# ///

import argparse
import asyncio
import csv
import json
import os
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Optional

from ray.job_submission import JobStatus, JobSubmissionClient

# We impose a 5min timeout here
# If any job does *not* finish in 5min, then we cancel it and mark the question as a "DNF" (did-not-finish).
TIMEOUT_S = 60 * 5


def parse_env_var_str(env_var_str: str) -> dict:
    iter = map(
        lambda s: s.strip().split("="),
        filter(lambda s: s, env_var_str.split(",")),
    )
    return {k: v for k, v in iter}


async def print_logs(logs):
    async for lines in logs:
        print(lines, end="")


async def wait_on_job(logs, timeout_s) -> bool:
    try:
        await asyncio.wait_for(print_logs(logs), timeout=timeout_s)
        return False
    except asyncio.exceptions.TimeoutError:
        return True


@dataclass
class Result:
    arguments: str
    duration: timedelta
    error_msg: Optional[str]


def submit_job(
    working_dir: Path,
    entrypoint_script: str,
    entrypoint_args: str,
    env_vars: str,
):
    if "GHA_OUTPUT_DIR" not in os.environ:
        raise RuntimeError("Output directory environment variable not found; don't know where to store outputs")
    output_dir = Path(os.environ["GHA_OUTPUT_DIR"])
    output_dir.mkdir(exist_ok=True, parents=True)

    env_vars_dict = parse_env_var_str(env_vars)

    client = JobSubmissionClient(address="http://localhost:8265")

    if entrypoint_args.startswith("[") and entrypoint_args.endswith("]"):
        # this is a json-encoded list of strings; parse accordingly
        list_of_entrypoint_args: list[str] = json.loads(entrypoint_args)
    else:
        list_of_entrypoint_args: list[str] = [entrypoint_args]

    results = []

    for args in list_of_entrypoint_args:
        entrypoint = f"DAFT_RUNNER=ray python {entrypoint_script} {args}"
        print(f"{entrypoint=}")
        start = datetime.now()
        job_id = client.submit_job(
            entrypoint=entrypoint,
            runtime_env={
                "working_dir": working_dir,
                "env_vars": env_vars_dict,
            },
        )

        timed_out = asyncio.run(wait_on_job(client.tail_job_logs(job_id), timeout_s=TIMEOUT_S))

        status = client.get_job_status(job_id)
        end = datetime.now()
        duration = end - start
        error_msg = None
        if status != JobStatus.SUCCEEDED:
            if timed_out:
                error_msg = f"Job exceeded {TIMEOUT_S} second(s)"
            else:
                job_info = client.get_job_info(job_id)
                error_msg = job_info.message

        result = Result(arguments=args, duration=duration, error_msg=error_msg)
        results.append(result)

    output_file = output_dir / "out.csv"
    with open(output_file, mode="w", newline="") as csv_file:
        writer = csv.DictWriter(csv_file, fieldnames=results[0].__dataclass_fields__.keys())
        writer.writeheader()
        for result in results:
            writer.writerow(asdict(result))


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--working-dir", type=Path, required=True)
    parser.add_argument("--entrypoint-script", type=str, required=True)
    parser.add_argument("--entrypoint-args", type=str, required=True)
    parser.add_argument("--env-vars", type=str, required=True)

    args = parser.parse_args()

    if not (args.working_dir.exists() and args.working_dir.is_dir()):
        raise ValueError("The working-dir must exist and be a directory")

    entrypoint: Path = args.working_dir / args.entrypoint_script
    if not (entrypoint.exists() and entrypoint.is_file()):
        raise ValueError("The entrypoint script must exist and be a file")

    submit_job(
        working_dir=args.working_dir,
        entrypoint_script=args.entrypoint_script,
        entrypoint_args=args.entrypoint_args,
        env_vars=args.env_vars,
    )



from __future__ import annotations

import contextlib
import datetime
import os
import tempfile
import uuid

import numpy as np
import pyarrow as pa
import pyarrow.parquet as papq
import pytest

import daft
from daft.datatype import DataType, TimeUnit
from daft.expressions import col
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition

from ..integration.io.conftest import minio_create_bucket

PYARROW_GE_11_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (11, 0, 0)
PYARROW_GE_13_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (13, 0, 0)


###
# Test Parquet Int96 timestamps
###


@contextlib.contextmanager
def _parquet_write_helper(data: pa.Table, row_group_size: int | None = None, papq_write_table_kwargs: dict = {}):
    with tempfile.TemporaryDirectory() as directory_name:
        file = os.path.join(directory_name, "tempfile")
        papq.write_table(data, file, row_group_size=row_group_size, **papq_write_table_kwargs)
        yield file


@pytest.mark.parametrize("use_deprecated_int96_timestamps", [True, False])
def test_parquet_read_int96_timestamps(use_deprecated_int96_timestamps):
    data = {
        "timestamp_ms": pa.array([1, 2, 3], pa.timestamp("ms")),
        "timestamp_us": pa.array([1, 2, 3], pa.timestamp("us")),
    }
    schema = [
        ("timestamp_ms", DataType.timestamp(TimeUnit.ms())),
        ("timestamp_us", DataType.timestamp(TimeUnit.us())),
    ]
    # int64 timestamps cannot support nanosecond resolutions
    if use_deprecated_int96_timestamps:
        data["timestamp_ns"] = pa.array([1, 2, 3], pa.timestamp("ns"))
        schema.append(("timestamp_ns", DataType.timestamp(TimeUnit.ns())))

    papq_write_table_kwargs = {
        "use_deprecated_int96_timestamps": use_deprecated_int96_timestamps,
        "coerce_timestamps": "us" if not use_deprecated_int96_timestamps else None,
    }
    if PYARROW_GE_11_0_0:
        papq_write_table_kwargs["store_schema"] = False

    with _parquet_write_helper(
        pa.Table.from_pydict(data),
        papq_write_table_kwargs=papq_write_table_kwargs,
    ) as f:
        expected = MicroPartition.from_pydict(data)
        df = daft.read_parquet(f, schema={k: v for k, v in schema})
        assert df.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{df.to_arrow()}"


@pytest.mark.parametrize("coerce_to", [TimeUnit.ms(), TimeUnit.us()])
def test_parquet_read_int96_timestamps_overflow(coerce_to):
    # NOTE: datetime.datetime(3000, 1, 1) and datetime.datetime(1000, 1, 1) cannot be represented by our timestamp64(nanosecond)
    # type. However they can be written to Parquet's INT96 type. Here we test that a round-trip is possible if provided with
    # the appropriate flags.
    data = {
        "timestamp": pa.array(
            [datetime.datetime(1000, 1, 1), datetime.datetime(2000, 1, 1), datetime.datetime(3000, 1, 1)],
            pa.timestamp(str(coerce_to)),
        ),
    }

    papq_write_table_kwargs = {
        "use_deprecated_int96_timestamps": True,
    }
    if PYARROW_GE_11_0_0:
        papq_write_table_kwargs["store_schema"] = False

    with _parquet_write_helper(
        pa.Table.from_pydict(data),
        papq_write_table_kwargs=papq_write_table_kwargs,
    ) as f:
        expected = MicroPartition.from_pydict(data)
        df = daft.read_parquet(f, coerce_int96_timestamp_unit=coerce_to)

        assert df.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{df}"


@pytest.mark.parametrize("coerce_to", [TimeUnit.ms(), TimeUnit.us()])
@pytest.mark.parametrize("store_schema", [True, False])
def test_parquet_read_int96_timestamps_schema_inference(coerce_to, store_schema):
    dt = datetime.datetime(2000, 1, 1)
    ns_ts_array = pa.array(
        [dt, dt, dt],
        pa.timestamp("ns"),
    )
    data = {
        "timestamp": ns_ts_array,
        "nested_timestamp": pa.array([[dt], [dt], [dt]], type=pa.list_(pa.timestamp("ns"))),
        "struct_timestamp": pa.array([{"foo": dt} for _ in range(3)], type=pa.struct({"foo": pa.timestamp("ns")})),
        "struct_nested_timestamp": pa.array(
            [{"foo": [dt]} for _ in range(3)], type=pa.struct({"foo": pa.list_(pa.timestamp("ns"))})
        ),
        "map_timestamp": pa.array([[("foo", dt)] for _ in range(3)], type=pa.map_(pa.string(), pa.timestamp("ns"))),
    }
    schema = [
        ("timestamp", DataType.timestamp(coerce_to)),
        ("nested_timestamp", DataType.list(DataType.timestamp(coerce_to))),
        ("struct_timestamp", DataType.struct({"foo": DataType.timestamp(coerce_to)})),
        ("struct_nested_timestamp", DataType.struct({"foo": DataType.list(DataType.timestamp(coerce_to))})),
        ("map_timestamp", DataType.map(DataType.string(), DataType.timestamp(coerce_to))),
    ]
    expected = Schema._from_field_name_and_types(schema)

    papq_write_table_kwargs = {
        "use_deprecated_int96_timestamps": True,
    }
    if PYARROW_GE_11_0_0:
        papq_write_table_kwargs["store_schema"] = store_schema

    with _parquet_write_helper(
        pa.Table.from_pydict(data),
        papq_write_table_kwargs=papq_write_table_kwargs,
    ) as f:
        schema = daft.read_parquet(f, coerce_int96_timestamp_unit=coerce_to).schema()
        assert schema == expected, f"Expected:\n{expected}\n\nReceived:\n{schema}"


def test_row_groups():
    path = ["tests/assets/parquet-data/mvp.parquet"]

    df = daft.read_parquet(path).collect()
    assert df.count_rows() == 100
    df = daft.read_parquet(path, row_groups=[[0, 1]]).collect()
    assert df.count_rows() == 20


# Test fix for issue #2537.
# This issue arose when the last row of a top-level column has a leaf field with values that span
# more than one data page.
@pytest.mark.integration()
@pytest.mark.parametrize("chunk_size", [5, 1024, 2048, 4096])
def test_parquet_rows_cross_page_boundaries(tmpdir, minio_io_config, chunk_size):
    int64_min = -(2**63)
    int64_max = 2**63 - 1

    def get_int_data_and_type(num_rows, repeat_int):
        data = [[{"field1": 1, "field2": 2}]]
        random_int_array = np.random.randint(int64_min, int64_max, size=(num_rows, repeat_int))
        data.extend([{"field1": int(field1), "field2": 2} for field1 in row] for row in random_int_array)
        data_type = pa.large_list(pa.struct([("field1", pa.int64()), ("field2", pa.int64())]))
        return data, data_type

    def get_string_data_and_type(num_rows, str_len, repeat_str):
        data = [[{"field1": "a", "field2": "b"}]]

        random_ascii_arrays = np.random.randint(32, 127, size=(num_rows, repeat_str, str_len), dtype=np.uint8)
        random_str_arrays = np.apply_along_axis(lambda x: "".join(map(chr, x)), 2, random_ascii_arrays)
        data.extend([{"field1": random_str, "field2": "b"} for random_str in row] for row in random_str_arrays)
        assert len(data) == num_rows + 1
        for i in range(1, num_rows + 1):
            assert len(data[i]) == repeat_str
            for j in range(repeat_str):
                assert len(data[i][j]["field1"]) == str_len

        data_type = pa.large_list(pa.struct([("field1", pa.large_string()), ("field2", pa.large_string())]))
        return data, data_type

    def get_dictionary_data_and_type(num_rows, str_len, repeat_str):
        data, _ = get_string_data_and_type(num_rows, str_len, repeat_str)
        data_type = pa.large_list(
            pa.struct(
                [
                    ("field1", pa.dictionary(pa.int32(), pa.string())),
                    ("field2", pa.large_string()),
                ]
            )
        )
        return data, data_type

    def compare_before_and_after(before, after):
        is_arrow = isinstance(before, pa.Table)
        has_dict = (
            pa.types.is_dictionary(before.schema.field("nested_col").type.field(0).type.field("field1").type)
            if is_arrow
            else False
        )

        # Test various combinations of limits, shows, and collects.
        after.limit(5).show()
        after.show()
        after.show(10)
        after = after.sort(col("_index"))
        if is_arrow:
            if has_dict:
                # Compare using to_pydict if there is a dictionary column because Daft does not support dictionary columns.
                assert before.to_pydict() == after.to_pydict()
            else:
                assert before == after.to_arrow()
        else:
            assert before.to_arrow() == after.to_arrow()
        after_limit_50 = after.limit(50)
        after_limit_2050 = after.limit(2050)  # Test a limit beyond the default chunk size (2048).
        if is_arrow:
            pd_table = before.to_pandas().explode("nested_col")
            assert [pd_table.count().get("nested_col")] == [
                x["count"] for x in after.explode(col("nested_col")).count().collect()
            ]
            before_limit_50 = before.take(list(range(min(before.num_rows, 50))))
            before_limit_2050 = before.take(list(range(min(before.num_rows, 2050))))
            if has_dict:
                # Compare using to_pydict if there is a dictionary column because Daft does not support dictionary columns.
                assert before_limit_50.to_pydict() == after_limit_50.to_pydict()
                assert before_limit_2050.to_pydict() == after_limit_2050.to_pydict()
            else:
                assert before_limit_50 == after_limit_50.to_arrow()
                assert before_limit_2050 == after_limit_2050.to_arrow()
            pd_table = before_limit_50.to_pandas().explode("nested_col")
            assert [pd_table.count().get("nested_col")] == [
                x["count"] for x in after_limit_50.explode(col("nested_col")).count().collect()
            ]
            pd_table = before_limit_2050.to_pandas().explode("nested_col")
            assert [pd_table.count().get("nested_col")] == [
                x["count"] for x in after_limit_2050.explode(col("nested_col")).count().collect()
            ]
        else:
            assert [x for x in before.explode(col("nested_col")).count().collect()] == [
                x for x in after.explode(col("nested_col")).count().collect()
            ]
            before_limit_50 = before.limit(50)
            before_limit_2050 = before.limit(2050)
            assert before_limit_50.to_arrow() == after_limit_50.to_arrow()
            assert [x for x in before_limit_50.explode(col("nested_col")).count().collect()] == [
                x for x in after_limit_50.explode(col("nested_col")).count().collect()
            ]
            assert before_limit_2050.to_arrow() == after_limit_2050.to_arrow()
            assert [x for x in before_limit_2050.explode(col("nested_col")).count().collect()] == [
                x for x in after_limit_2050.explode(col("nested_col")).count().collect()
            ]

    def test_parquet_helper(data_and_type, use_daft_writer):
        data, data_type = data_and_type
        index_data = [x for x in range(0, len(data))]
        file_path = f"{tmpdir}/{uuid.uuid4()!s}.parquet"

        # Test Daft roundtrip. Daft does not support the dictionary logical type, hence we skip
        # writing with Daft for this type.
        if use_daft_writer:
            before = daft.from_pydict(
                {"nested_col": pa.array(data, type=data_type), "_index": pa.array(index_data, type=pa.int64())}
            )
            before = before.sort(col("_index"))
            before.write_parquet(file_path)
            after = daft.read_parquet(file_path, _chunk_size=chunk_size)
            compare_before_and_after(before, after)
            # Test reads from S3.
            bucket_name = "my-bucket"
            s3_path = f"s3://{bucket_name}/my-folder"
            with minio_create_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
                before.write_parquet(s3_path, io_config=minio_io_config)
                after = daft.read_parquet(s3_path, io_config=minio_io_config, _chunk_size=chunk_size)
                compare_before_and_after(before, after)

        # Test Arrow write with Daft read.
        file_path = f"{tmpdir}/{uuid.uuid4()!s}.parquet"
        before = pa.Table.from_arrays(
            [pa.array(data, type=data_type), pa.array(index_data, type=pa.int64())], names=["nested_col", "_index"]
        )
        before = before.sort_by("_index")
        write_options = papq.ParquetWriter(
            file_path,
            before.schema,
            compression="SNAPPY",
            use_dictionary={"string_column": True},
            data_page_size=1024 * 1024,
        )
        with write_options as writer:
            writer.write_table(before)
        after = daft.read_parquet(file_path, _chunk_size=chunk_size)
        compare_before_and_after(before, after)

    # The normal case where the last row `nested.field1` is contained within a single data page.
    # Data page has 131071 items.
    test_parquet_helper(get_int_data_and_type(65535, 2), True)
    # Data page has 1023 items.
    test_parquet_helper(get_string_data_and_type(511, 3000, 2), True)
    test_parquet_helper(get_dictionary_data_and_type(511, 3000, 2), False)
    # Data pages have 10240 and 9761 items. This could cause .show() or .limit() to fail if we
    # incorrectly account for the number of values to read.
    test_parquet_helper(get_string_data_and_type(100, 100, 200), True)

    # Cases where the last row of `nested.field1` has items that span two data pages.
    # Data pages have 131072 and 1 items.
    test_parquet_helper(get_int_data_and_type(65536, 2), True)
    # Data pages have 1024 and 1 items.
    test_parquet_helper(get_string_data_and_type(512, 3000, 2), True)
    test_parquet_helper(get_dictionary_data_and_type(512, 3000, 2), False)
    # Data pages have 131072, 131072, and 1 items.
    test_parquet_helper(get_int_data_and_type(131072, 2), True)
    # Data pages has 1024, 1024, and 1 items.
    test_parquet_helper(get_string_data_and_type(1024, 3000, 2), True)
    test_parquet_helper(get_dictionary_data_and_type(1024, 3000, 2), False)

    # Cases where the last row of `nested.field1` has items that span multiple data pages.
    # Data pages has 131072, 131072, and 1 items.
    test_parquet_helper(get_int_data_and_type(1, 262144), True)
    # Data pages has 1024, 1024, 1024, 1024, and 1 items.
    test_parquet_helper(get_string_data_and_type(1, 3000, 3072), True)
    test_parquet_helper(get_dictionary_data_and_type(1, 3000, 3072), False)

    # Cases where the list sizes are 1. This also simulates the case where there is no list in the
    # schema. We encountered a bug where if page size aligns with chunk size (typically 2048 for
    # non-local reads), then upon checking the next page for more values, we would read
    # more rows than the chunk size, and the `rows read == chunk size` check would not be true until
    # we've read all values in the page. This could repeat for every subsequent data page.

    # Here we test various scenarios where the number of values in a data page are various multiples
    # and denominators of the parameterized chunk size.

    # One column uses a single dictionary-encoded data page, and the other contains data pages with
    # 1024 values each.
    test_parquet_helper(get_string_data_and_type(4096, 3000, 1), True)
    # One column uses a single dictionary-encoded data page, and the other contains data pages with
    # 2048 values each.
    test_parquet_helper(get_string_data_and_type(8192, 1000, 1), True)
    # One column uses a single dictionary-encoded data page, and the other contains data pages with
    # 4096 values each.
    test_parquet_helper(get_string_data_and_type(8192, 300, 1), True)


@pytest.mark.integration()
def test_parquet_limits_across_row_groups(tmpdir, minio_io_config):
    test_row_group_size = 1024
    daft_execution_config = daft.context.get_context().daft_execution_config
    default_row_group_size = daft_execution_config.parquet_target_row_group_size
    int_array = np.full(shape=4096, fill_value=3, dtype=np.int32)
    before = daft.from_pydict({"col": pa.array(int_array, type=pa.int32())})
    file_path = f"{tmpdir}/{uuid.uuid4()!s}.parquet"
    # Decrease the target row group size before writing the parquet file.
    daft.set_execution_config(parquet_target_row_group_size=test_row_group_size)
    before.write_parquet(file_path)
    assert (
        before.limit(test_row_group_size + 10).to_arrow()
        == daft.read_parquet(file_path).limit(test_row_group_size + 10).to_arrow()
    )
    assert (
        before.limit(test_row_group_size * 2).to_arrow()
        == daft.read_parquet(file_path).limit(test_row_group_size * 2).to_arrow()
    )

    bucket_name = "my-bucket"
    s3_path = f"s3://{bucket_name}/my-folder"
    with minio_create_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
        before.write_parquet(s3_path, io_config=minio_io_config)
        assert (
            before.limit(test_row_group_size + 10).to_arrow()
            == daft.read_parquet(s3_path, io_config=minio_io_config).limit(test_row_group_size + 10).to_arrow()
        )
        assert (
            before.limit(test_row_group_size * 2).to_arrow()
            == daft.read_parquet(s3_path, io_config=minio_io_config).limit(test_row_group_size * 2).to_arrow()
        )
    # Reset the target row group size.
    daft.set_execution_config(parquet_target_row_group_size=default_row_group_size)


@pytest.mark.parametrize("optional_outer_struct", [True, False])
@pytest.mark.parametrize("optional_inner_struct", [True, False])
def test_parquet_nested_optional_or_required_fields(tmpdir, optional_outer_struct, optional_inner_struct):
    schema = pa.schema(
        [
            pa.field(
                "outer_struct_field",
                pa.struct(
                    [
                        pa.field(
                            "inner_struct_field",
                            pa.struct(
                                [
                                    pa.field("optional_field_str", pa.string()),
                                    pa.field("optional_field_binary", pa.binary()),
                                    pa.field("optional_field_int", pa.int32()),
                                    pa.field("optional_field_bool", pa.bool_()),
                                    pa.field("required_field_str", pa.string(), nullable=False),
                                    pa.field("required_field_binary", pa.binary(), nullable=False),
                                    pa.field("required_field_int", pa.int32(), nullable=False),
                                    pa.field("required_field_bool", pa.bool_(), nullable=False),
                                ]
                            ),
                            nullable=optional_inner_struct,
                        )
                    ]
                ),
                nullable=optional_outer_struct,
            )
        ]
    )
    num_records = 8192
    data = [
        {
            "outer_struct_field": None
            if optional_outer_struct and i % 4 == 0
            else {
                "inner_struct_field": None
                if optional_inner_struct and i % 5 == 0
                else {
                    "optional_field_str": f"string_{i}" if i % 3 != 0 else None,
                    "optional_field_binary": f"binary_{i}".encode() if i % 5 != 0 else None,
                    "optional_field_int": i if i % 7 != 0 else None,
                    "optional_field_bool": bool(i % 3) if i % 11 != 0 else None,
                    "required_field_str": f"string_{i}",
                    "required_field_binary": f"binary_{i}".encode(),
                    "required_field_int": i,
                    "required_field_bool": bool(i % 3),
                }
            }
        }
        for i in range(num_records)
    ]
    expected = pa.Table.from_pylist(data, schema=schema)
    output_file = f"{tmpdir}/{uuid.uuid4()!s}.parquet"
    papq.write_table(expected, output_file)
    expected = MicroPartition.from_arrow(expected)
    df = daft.read_parquet(output_file)
    assert df.to_arrow() == expected.to_arrow(), f"Expected:\n{expected.to_arrow()}\n\nReceived:\n{df.to_arrow()}"



import pytest

import daft
from daft.sql import SQLCatalog


@pytest.mark.parametrize(
    "query,expected",
    [
        ("SELECT * FROM df1 WHERE val <=> 20", {"id": [2], "val": [20]}),
        ("SELECT * FROM df1 WHERE val <=> NULL", {"id": [3], "val": [None]}),
        (
            "SELECT df1.id, df1.val, df2.score FROM df1 JOIN df2 ON df1.id <=> df2.id",
            {"id": [1, 2, None], "val": [10, 20, 40], "score": [0.1, 0.2, 0.3]},
        ),
        (
            "SELECT * FROM df1 WHERE val <=> 10 OR val <=> NULL",
            {"id": [1, 3], "val": [10, None]},  # Matches both 10 and NULL values
        ),
    ],
)
def test_null_safe_equals_basic(query, expected):
    """Test basic null-safe equality operator (<=>)."""
    df1 = daft.from_pydict({"id": [1, 2, 3, None], "val": [10, 20, None, 40]})
    df2 = daft.from_pydict({"id": [1, 2, None, 4], "score": [0.1, 0.2, 0.3, 0.4]})

    catalog = SQLCatalog({"df1": df1, "df2": df2})
    result = daft.sql(query, catalog).to_pydict()
    assert result == expected


@pytest.mark.parametrize(
    "query,expected",
    [
        ("SELECT * FROM df WHERE NOT (val <=> NULL)", {"id": [1, 3, None], "val": [10, 30, 40]}),
        ("SELECT * FROM df WHERE val <=> 10 OR id > 2", {"id": [1, 3], "val": [10, 30]}),
        (
            "SELECT *, CASE WHEN val <=> NULL THEN 'is_null' ELSE 'not_null' END as val_status FROM df",
            {
                "id": [1, 2, 3, None, None],
                "val": [10, None, 30, 40, None],
                "val_status": ["not_null", "is_null", "not_null", "not_null", "is_null"],
            },
        ),
    ],
)
def test_null_safe_equals_complex(query, expected):
    """Test complex expressions using null-safe equality."""
    df = daft.from_pydict({"id": [1, 2, 3, None, None], "val": [10, None, 30, 40, None]})

    catalog = SQLCatalog({"df": df})
    result = daft.sql(query, catalog).to_pydict()
    assert result == expected


@pytest.mark.parametrize(
    "query,expected",
    [
        (
            "SELECT * FROM df WHERE int_val <=> 1",
            {"int_val": [1], "str_val": ["a"], "bool_val": [True], "float_val": [1.1]},
        ),
        (
            "SELECT * FROM df WHERE str_val <=> 'c'",
            {"int_val": [3], "str_val": ["c"], "bool_val": [False], "float_val": [3.3]},
        ),
        (
            "SELECT * FROM df WHERE bool_val <=> false",
            {"int_val": [3], "str_val": ["c"], "bool_val": [False], "float_val": [3.3]},
        ),
        (
            "SELECT * FROM df WHERE float_val <=> 1.1",
            {"int_val": [1], "str_val": ["a"], "bool_val": [True], "float_val": [1.1]},
        ),
        (
            "SELECT * FROM df WHERE int_val <=> NULL",
            {"int_val": [None], "str_val": [None], "bool_val": [None], "float_val": [None]},
        ),
        (
            "SELECT * FROM df WHERE str_val <=> NULL",
            {"int_val": [None], "str_val": [None], "bool_val": [None], "float_val": [None]},
        ),
        (
            "SELECT * FROM df WHERE bool_val <=> NULL",
            {"int_val": [None], "str_val": [None], "bool_val": [None], "float_val": [None]},
        ),
        (
            "SELECT * FROM df WHERE float_val <=> NULL",
            {"int_val": [None], "str_val": [None], "bool_val": [None], "float_val": [None]},
        ),
    ],
)
def test_null_safe_equals_types(query, expected):
    """Test null-safe equality with different data types."""
    df = daft.from_pydict(
        {
            "int_val": [1, None, 3],
            "str_val": ["a", None, "c"],
            "bool_val": [True, None, False],
            "float_val": [1.1, None, 3.3],
        }
    )

    catalog = SQLCatalog({"df": df})
    result = daft.sql(query, catalog).to_pydict()
    assert result == expected



from __future__ import annotations

import copy
from typing import Dict, List

import pytest

from daft.datatype import DataType

daft_int_types = [
    DataType.int8(),
    DataType.int16(),
    DataType.int32(),
    DataType.int64(),
    DataType.uint8(),
    DataType.uint16(),
    DataType.uint32(),
    DataType.uint64(),
]

daft_numeric_types = daft_int_types + [DataType.float32(), DataType.float64()]
daft_string_types = [DataType.string()]
daft_nonnull_types = (
    daft_numeric_types
    + daft_string_types
    + [DataType.bool(), DataType.binary(), DataType.fixed_size_binary(1), DataType.date()]
)


@pytest.mark.parametrize("dtype", daft_nonnull_types)
def test_datatype_pickling(dtype) -> None:
    copy_dtype = copy.deepcopy(dtype)
    assert copy_dtype == dtype


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        (str, DataType.string()),
        (int, DataType.int64()),
        (float, DataType.float64()),
        (bytes, DataType.binary()),
        (object, DataType.python()),
        (
            {"foo": str, "bar": int},
            DataType.struct({"foo": DataType.string(), "bar": DataType.int64()}),
        ),
    ],
)
def test_datatype_parsing(source, expected):
    assert DataType._infer_type(source) == expected


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        # These tests must be run in later version of Python that allow for subscripting of types
        (list[str], DataType.list(DataType.string())),
        (dict[str, int], DataType.map(DataType.string(), DataType.int64())),
        (
            {"foo": list[str], "bar": int},
            DataType.struct({"foo": DataType.list(DataType.string()), "bar": DataType.int64()}),
        ),
        (list[list[str]], DataType.list(DataType.list(DataType.string()))),
    ],
)
def test_subscripted_datatype_parsing(source, expected):
    assert DataType._infer_type(source) == expected


@pytest.mark.parametrize(
    ["source", "expected"],
    [
        # These tests must be run in later version of Python that allow for subscripting of types
        (List[str], DataType.list(DataType.string())),
        (Dict[str, int], DataType.map(DataType.string(), DataType.int64())),
        (
            {"foo": List[str], "bar": int},
            DataType.struct({"foo": DataType.list(DataType.string()), "bar": DataType.int64()}),
        ),
        (List[List[str]], DataType.list(DataType.list(DataType.string()))),
    ],
)
def test_legacy_subscripted_datatype_parsing(source, expected):
    assert DataType._infer_type(source) == expected



from typing import TYPE_CHECKING, Dict, Iterator, List, Optional, Tuple

from daft.context import get_context
from daft.daft import IOConfig
from daft.datatype import DataType
from daft.dependencies import pa
from daft.io.common import _get_schema_from_dict
from daft.recordbatch.micropartition import MicroPartition
from daft.recordbatch.partitioning import PartitionedTable, partition_strings_to_path

if TYPE_CHECKING:
    from deltalake.writer import AddAction


def sanitize_table_for_deltalake(
    table: MicroPartition, large_dtypes: bool, partition_keys: Optional[List[str]] = None
) -> pa.Table:
    from deltalake.schema import _convert_pa_schema_to_delta

    from daft.io._deltalake import large_dtypes_kwargs

    arrow_table = table.to_arrow()

    # Remove partition keys from the table since they are already encoded as keys
    if partition_keys is not None:
        arrow_table = arrow_table.drop_columns(partition_keys)

    arrow_batch = _convert_pa_schema_to_delta(arrow_table.schema, **large_dtypes_kwargs(large_dtypes))
    return arrow_table.cast(arrow_batch)


def partitioned_table_to_deltalake_iter(
    partitioned: PartitionedTable, large_dtypes: bool
) -> Iterator[Tuple[pa.Table, str, Dict[str, Optional[str]]]]:
    """Iterates over partitions, yielding each partition as an Arrow table, along with their respective paths and partition values."""
    partition_values = partitioned.partition_values()

    if partition_values:
        partition_keys = partition_values.column_names()
        partition_strings = partitioned.partition_values_str()
        assert partition_strings is not None

        for part_table, part_strs in zip(partitioned.partitions(), partition_strings.to_pylist()):
            part_path = partition_strings_to_path("", part_strs)
            converted_arrow_table = sanitize_table_for_deltalake(part_table, large_dtypes, partition_keys)
            yield converted_arrow_table, part_path, part_strs
    else:
        converted_arrow_table = sanitize_table_for_deltalake(partitioned.table, large_dtypes)
        yield converted_arrow_table, "/", {}


def make_deltalake_add_action(
    path,
    metadata,
    size,
    partition_values,
):
    import json
    from datetime import datetime

    import deltalake
    from deltalake.writer import (
        AddAction,
        DeltaJSONEncoder,
        get_file_stats_from_metadata,
    )
    from packaging.version import parse

    # added to get_file_stats_from_metadata in deltalake v0.17.4: non-optional "num_indexed_cols" and "columns_to_collect_stats" arguments
    # https://github.com/delta-io/delta-rs/blob/353e08be0202c45334dcdceee65a8679f35de710/python/deltalake/writer.py#L725
    if parse(deltalake.__version__) < parse("0.17.4"):
        file_stats_args = {}
    else:
        file_stats_args = {"num_indexed_cols": -1, "columns_to_collect_stats": None}

    stats = get_file_stats_from_metadata(metadata, **file_stats_args)

    # remove leading slash
    path = path[1:] if path.startswith("/") else path
    return AddAction(
        path,
        size,
        partition_values,
        int(datetime.now().timestamp() * 1000),
        True,
        json.dumps(stats, cls=DeltaJSONEncoder),
    )


def make_deltalake_fs(path: str, io_config: Optional[IOConfig] = None):
    from deltalake.writer import DeltaStorageHandler
    from pyarrow.fs import PyFileSystem

    from daft.io.object_store_options import io_config_to_storage_options

    io_config = get_context().daft_planning_config.default_io_config if io_config is None else io_config
    storage_options = io_config_to_storage_options(io_config, path)
    return PyFileSystem(DeltaStorageHandler(path, storage_options))


class DeltaLakeWriteVisitors:
    class FileVisitor:
        def __init__(
            self,
            parent: "DeltaLakeWriteVisitors",
            partition_values: Dict[str, Optional[str]],
        ):
            self.parent = parent
            self.partition_values = partition_values

        def __call__(self, written_file):
            from daft.utils import get_arrow_version

            # PyArrow added support for size in 9.0.0
            if get_arrow_version() >= (9, 0, 0):
                size = written_file.size
            elif self.parent.fs is not None:
                size = self.parent.fs.get_file_info([written_file.path])[0].size
            else:
                size = 0

            add_action = make_deltalake_add_action(
                written_file.path, written_file.metadata, size, self.partition_values
            )

            self.parent.add_actions.append(add_action)

    def __init__(self, fs: pa.fs.FileSystem):
        self.add_actions: List[AddAction] = []
        self.fs = fs

    def visitor(self, partition_values: Dict[str, Optional[str]]) -> "DeltaLakeWriteVisitors.FileVisitor":
        return self.FileVisitor(self, partition_values)

    def to_metadata(self) -> MicroPartition:
        col_name = "add_action"
        if len(self.add_actions) == 0:
            return MicroPartition.empty(_get_schema_from_dict({col_name: DataType.python()}))
        return MicroPartition.from_pydict({col_name: self.add_actions})



from __future__ import annotations

import sys

import pytest

import daft
from benchmarking.tpch import answers, data_generation

# Hardcode scale factor to 200M for local testing
SCALE_FACTOR = 0.2

if sys.platform == "win32":
    pytest.skip(allow_module_level=True)


@pytest.fixture(scope="module")
def get_df_from_sql(gen_tpch):
    _, sqlite_db_file_path = gen_tpch

    def _get_df(tbl_name: str):
        df = daft.read_sql(f"SELECT * FROM {tbl_name}", f"sqlite://{sqlite_db_file_path}")
        df = df.select(
            *[
                daft.col(autoname).alias(colname)
                for autoname, colname in zip(df.column_names, data_generation.SCHEMA[tbl_name])
            ]
        )
        return df

    return _get_df


def test_tpch_q1(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q1(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 1, tmp_path)


def test_tpch_q2(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q2(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 2, tmp_path)


def test_tpch_q3(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q3(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 3, tmp_path)


def test_tpch_q4(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q4(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 4, tmp_path)


def test_tpch_q5(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q5(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 5, tmp_path)


def test_tpch_q6(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q6(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 6, tmp_path)


def test_tpch_q7(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q7(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 7, tmp_path)


def test_tpch_q8(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q8(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 8, tmp_path)


def test_tpch_q9(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q9(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 9, tmp_path)


def test_tpch_q10(tmp_path, check_answer, get_df_from_sql):
    daft_df = answers.q10(get_df_from_sql)
    daft_pd_df = daft_df.to_pandas()
    check_answer(daft_pd_df, 10, tmp_path)



#!/usr/bin/env python3

"""This script generates the files needed to implement Python's Simple repository API (https://packaging.python.org/en/latest/specifications/simple-repository-api/#simple-repository-api).

Given a S3 bucket with these files:

bucket/
- foo-1.0-none-any.whl
- foo-1.1-none-any.whl
- bar-1.0-none-any.whl

It will generate these files in `dist/indices/`:
- index.html
- foo/index.html
- bar/index.html

Then, put these files in the same bucket with a Cloudfront distribution configured to serve via HTTPS and serve index.html for directory paths.
"""

import os
import sys
import urllib.parse

import boto3
from packaging.utils import parse_wheel_filename


def write_file_ensure_dir(filename, s):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, "w") as f:
        f.write(s)


def generate_root_index(prefix, pkg_names):
    links = "\n".join([f'<a href="/{prefix}/{urllib.parse.quote(name)}/">{name}</a>' for name in pkg_names])

    return f"""<!DOCTYPE html>
<html>
  <body>
    {links}
  </body>
</html>"""


def generate_pkg_index(prefix, wheel_names):
    links = "\n".join([f'<a href="/{prefix}/{urllib.parse.quote(name)}">{name}</a>' for name in wheel_names])

    return f"""<!DOCTYPE html>
<html>
  <body>
    {links}
  </body>
</html>"""


def main():
    if len(sys.argv) != 2:
        print("Usage: python generate_simple_py_index.py s3://bucket")
        sys.exit(1)

    s3_prefix = "s3://"
    s3_url = sys.argv[1]
    assert s3_url.startswith(s3_prefix)
    (bucket, _, prefix) = s3_url.removeprefix(s3_prefix).removesuffix("/").partition("/")

    s3 = boto3.client("s3")
    paginator = s3.get_paginator("list_objects_v2")
    pages = paginator.paginate(Bucket=bucket, Prefix=prefix)

    pkg_map = {}
    for page in pages:
        for obj in page["Contents"]:
            if obj["Key"].endswith(".whl"):
                wheel_name = obj["Key"].removeprefix(prefix + "/")

                pkg_name, _, _, _ = parse_wheel_filename(wheel_name)
                if pkg_name not in pkg_map:
                    pkg_map[pkg_name] = []
                pkg_map[pkg_name].append(wheel_name)

    root_index = generate_root_index(prefix, pkg_map.keys())
    write_file_ensure_dir("dist/indices/index.html", root_index)

    for pkg_name, wheel_names in pkg_map.items():
        pkg_index = generate_pkg_index(prefix, wheel_names)
        write_file_ensure_dir(f"dist/indices/{pkg_name}/index.html", pkg_index)


if __name__ == "__main__":
    main()



"""An in-memory implementation for the daft catalog abstractions."""

from __future__ import annotations

from typing import TYPE_CHECKING

from daft.catalog import Catalog, Identifier, Table

if TYPE_CHECKING:
    from daft.dataframe.dataframe import DataFrame


class MemoryCatalog(Catalog):
    """An in-memory catalog is backed by a dictionary."""

    _tables: dict[str, Table]

    def __init__(self, tables: dict[str, Table]):
        self._tables = tables

    ###
    # list_*
    ###

    def list_tables(self, pattern: str | None = None) -> list[str]:
        if pattern is None:
            return list(self._tables.keys())
        return [path for path in self._tables.keys() if path.startswith(pattern)]

    ###
    # get_*
    ###

    def get_table(self, name: str | Identifier) -> Table:
        path = str(name)
        if path not in self._tables:
            raise ValueError(f"Table {path} does not exist.")
        return self._tables[path]


class MemoryTable(Table):
    """An in-memory table holds a reference to an existing dataframe."""

    _name: str
    _inner: DataFrame

    def __init__(self, name: str, inner: DataFrame):
        self._name = name
        self._inner = inner

    ###
    # DataFrame Methods
    ###

    def read(self) -> DataFrame:
        return self._inner



"""WARNING! These APIs are internal; please use Catalog.from_unity() and Table.from_unity()."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from daft.catalog import Catalog, Identifier, Table
from daft.unity_catalog import UnityCatalog as InnerCatalog  # noqa: TID253
from daft.unity_catalog import UnityCatalogTable as InnerTable  # noqa: TID253

if TYPE_CHECKING:
    from daft.dataframe import DataFrame


class UnityCatalog(Catalog):
    _inner: InnerCatalog

    def __init__(self, unity_catalog: InnerCatalog):
        """DEPRECATED: Please use `Catalog.from_unity`; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please prefer using `Catalog.from_unity` instead; version 0.5.0!",
            category=DeprecationWarning,
        )
        self._inner = unity_catalog

    @staticmethod
    def _from_obj(obj: object) -> UnityCatalog:
        """Returns an UnityCatalog instance if the given object can be adapted so."""
        if isinstance(obj, InnerCatalog):
            c = UnityCatalog.__new__(UnityCatalog)
            c._inner = obj
            return c
        raise ValueError(f"Unsupported unity catalog type: {type(obj)}")

    ###
    # get_*
    ###

    def get_table(self, ident: Identifier | str) -> UnityTable:
        if isinstance(ident, Identifier):
            ident = ".".join(ident)  # TODO unity qualified identifiers
        return UnityTable(self._inner.load_table(ident))

    ###
    # list_.*
    ###

    def list_tables(self, pattern: str | None = None) -> list[str]:
        if pattern is None or pattern == "":
            return [
                tbl
                for cat in self._inner.list_catalogs()
                for schema in self._inner.list_schemas(cat)
                for tbl in self._inner.list_tables(schema)
            ]
        num_namespaces = pattern.count(".")
        if num_namespaces == 0:
            catalog_name = pattern
            return [tbl for schema in self._inner.list_schemas(catalog_name) for tbl in self._inner.list_tables(schema)]
        elif num_namespaces == 1:
            schema_name = pattern
            return [tbl for tbl in self._inner.list_tables(schema_name)]
        else:
            raise ValueError(
                f"Unrecognized catalog name or schema name, expected a '.'-separated namespace but received: {pattern}"
            )


class UnityTable(Table):
    _inner: InnerTable

    def __init__(self, unity_table: InnerTable):
        """DEPRECATED: Please use `Table.from_unity`; version 0.5.0!"""
        warnings.warn(
            "This is deprecated and will be removed in daft >= 0.5.0, please prefer using `Table.from_unity` instead; version 0.5.0!",
            category=DeprecationWarning,
        )
        self._inner = unity_table

    @staticmethod
    def _from_obj(obj: object) -> UnityTable | None:
        """Returns a UnityTable if the given object can be adapted so."""
        if isinstance(obj, InnerTable):
            t = UnityTable.__new__(UnityTable)
            t._inner = obj
            return t
        raise ValueError(f"Unsupported unity table type: {type(obj)}")

    @staticmethod
    def _try_from(obj: object) -> UnityTable | None:
        """Returns an UnityTable if the given object can be adapted so."""
        if isinstance(obj, InnerTable):
            return UnityTable(obj)
        return None

    @property
    def inner(self) -> InnerTable:
        """Returns the inner unity table."""
        return self._inner

    def read(self) -> DataFrame:
        import daft

        return daft.read_deltalake(self._inner)



from __future__ import annotations

import contextlib
import csv
import os
import pathlib
import tempfile

import pytest

import daft
from daft.daft import CsvParseOptions
from daft.datatype import DataType
from daft.logical.schema import Schema
from daft.recordbatch import MicroPartition, recordbatch_io
from daft.runners.partitioning import TableParseCSVOptions, TableReadOptions
from daft.utils import get_arrow_version


def test_read_input(tmpdir):
    tmpdir = pathlib.Path(tmpdir)
    data = {"foo": [1, 2, 3]}
    with open(tmpdir / "file.csv", "w") as f:
        f.write("foo\n")
        f.write("1\n")
        f.write("2\n")
        f.write("3\n")

    schema = Schema._from_field_name_and_types([("foo", DataType.int64())])

    # Test pathlib, str and IO
    assert recordbatch_io.read_csv(tmpdir / "file.csv", schema=schema).to_pydict() == data
    assert recordbatch_io.read_csv(str(tmpdir / "file.csv"), schema=schema).to_pydict() == data


@contextlib.contextmanager
def _csv_write_helper(header: list[str] | None, data: list[list[str | None]], **kwargs):
    with tempfile.TemporaryDirectory() as directory_name:
        file = os.path.join(directory_name, "tempfile")
        with open(file, "w", newline="") as f:
            writer = csv.writer(f, **kwargs)
            if header is not None:
                writer.writerow(header)
            writer.writerows(data)
        yield file


@pytest.mark.parametrize(
    ["data", "expected_dtype"],
    [
        ("1", DataType.int64()),
        ("foo", DataType.string()),
        ("1.5", DataType.float64()),
        ("True", DataType.bool()),
    ],
)
def test_csv_infer_schema(data, expected_dtype):
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", data],
            ["2", data],
            ["3", None],
        ],
    ) as f:
        schema = Schema.from_csv(f)
        assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", expected_dtype)])


def test_csv_infer_schema_custom_delimiter():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", "1"],
            ["2", "2"],
            ["3", None],
        ],
        delimiter="|",
    ) as f:
        schema = Schema.from_csv(f, parse_options=CsvParseOptions(delimiter="|"))
        assert schema == Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])


def test_csv_infer_schema_no_header():
    with _csv_write_helper(
        header=None,
        data=[
            ["1", "1"],
            ["2", "2"],
            ["3", None],
        ],
    ) as f:
        schema = Schema.from_csv(f, parse_options=CsvParseOptions(has_header=False))
        fields = [("column_1", DataType.int64()), ("column_2", DataType.int64())]
        assert schema == Schema._from_field_name_and_types(fields)


@pytest.mark.parametrize(
    ["data", "expected_data_series"],
    [
        ["1", daft.Series.from_pylist([1, 1, None])],
        # NOTE: Empty gets parsed as "" instead of None for string fields
        ["foo", daft.Series.from_pylist(["foo", "foo", ""])],
        ["1.5", daft.Series.from_pylist([1.5, 1.5, None])],
        ("True", daft.Series.from_pylist([True, True, None])),
    ],
)
def test_csv_read_data(data, expected_data_series):
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", data],
            ["2", data],
            ["3", None],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types(
            [("id", DataType.int64()), ("data", expected_data_series.datatype())]
        )
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": expected_data_series,
            }
        )
        table = recordbatch_io.read_csv(f, schema)
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_csv_limit_rows():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", "1"],
            ["2", "2"],
            ["3", None],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2],
                "data": [1, 2],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            read_options=TableReadOptions(num_rows=2),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_csv_select_columns():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", "1"],
            ["2", "2"],
            ["3", None],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "data": [1, 2, None],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            read_options=TableReadOptions(column_names=["data"]),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_csv_custom_delimiter():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", "1"],
            ["2", "2"],
            ["3", None],
        ],
        delimiter="|",
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": [1, 2, None],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(delimiter="|"),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_csv_no_header():
    with _csv_write_helper(
        header=None,
        data=[
            ["1", "1"],
            ["2", "2"],
            ["3", None],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.int64())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": [1, 2, None],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(header_index=None),
            read_options=TableReadOptions(column_names=["id", "data"]),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_csv_custom_quote():
    with _csv_write_helper(
        header=["'id'", "'data'"],
        data=[
            ["1", "'aa'"],
            ["2", "aa"],
            ["3", "aa"],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.string())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": ["aa", "aa", "aa"],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(quote="'"),
        )

        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_custom_escape():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", 'a"a"a'],
            ["2", "aa"],
            ["3", "aa"],
        ],
        quotechar='"',
        escapechar="\\",
        doublequote=False,
        quoting=csv.QUOTE_ALL,
    ) as f:
        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.string())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2, 3],
                "data": ['a"a"a', "aa", "aa"],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(escape_char="\\", double_quote=False),
        )

        assert table.to_arrow() == expected.to_arrow(), f"Received:\n{table}\n\nExpected:\n{expected}"


def test_csv_read_data_custom_comment():
    with tempfile.TemporaryDirectory() as directory_name:
        file = os.path.join(directory_name, "tempfile")
        with open(file, "w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["id", "data"])
            writer.writerow(["1", "aa"])
            f.write("# comment line\n")
            writer.writerow(["3", "aa"])

        schema = Schema._from_field_name_and_types([("id", DataType.int64()), ("data", DataType.string())])
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 3],
                "data": ["aa", "aa"],
            }
        )
        # Skipping test for arrow < 7.0.0 as comments are not supported in pyarrow
        arrow_version = get_arrow_version()
        if arrow_version >= (7, 0, 0):
            table = recordbatch_io.read_csv(
                file,
                schema,
                csv_options=TableParseCSVOptions(comment="#"),
            )
            assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_variable_missing_columns():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1"],
            ["2", "2"],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types(
            [
                ("id", DataType.int64()),
                ("data", DataType.int64()),
            ]
        )
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2],
                "data": [None, 2],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(allow_variable_columns=True),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_variable_extra_columns():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["1", "1"],
            ["2", "2", "2"],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types(
            [
                ("id", DataType.int64()),
                ("data", DataType.int64()),
            ]
        )
        expected = MicroPartition.from_pydict(
            {
                "id": [1, 2],
                "data": [1, 2],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(allow_variable_columns=True),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"


def test_csv_read_data_variable_columns_with_non_matching_types():
    with _csv_write_helper(
        header=["id", "data"],
        data=[
            ["a", "b"],
            ["2", "2"],
        ],
    ) as f:
        schema = Schema._from_field_name_and_types(
            [
                ("id", DataType.int64()),
                ("data", DataType.int64()),
            ]
        )
        expected = MicroPartition.from_pydict(
            {
                "id": [None, 2],
                "data": [None, 2],
            }
        )
        table = recordbatch_io.read_csv(
            f,
            schema,
            csv_options=TableParseCSVOptions(allow_variable_columns=True),
        )
        assert table.to_arrow() == expected.to_arrow(), f"Expected:\n{expected}\n\nReceived:\n{table}"



import inspect
import os

from sphinx.ext.autosummary import Autosummary
from sphinx.util import logging

logger = logging.getLogger(__name__)


TOCTREE = "doc_gen/sql_funcs"
SQL_MODULE_NAME = "daft.sql._sql_funcs"

STUB_TEMPLATE = """
.. currentmodule:: None

.. autofunction:: {module_name}.{name}
"""


class SQLAutosummary(Autosummary):
    def run(self):
        func_names = get_sql_func_names()
        # Run the normal autosummary stuff, override self.content
        self.content = [f"~{SQL_MODULE_NAME}.{f}" for f in func_names]
        nodes = super().run()
        return nodes

    def get_sql_module_name(self):
        return self.arguments[0]


def get_sql_func_names():
    # Import the SQL functions module
    module = __import__(SQL_MODULE_NAME, fromlist=[""])

    names = []
    for name, obj in inspect.getmembers(module):
        if inspect.isfunction(obj) and not name.startswith("_"):
            names.append(name)

    return names


def generate_stub(name: str):
    """Generates a stub string for a SQL function."""
    stub = name + "\n"
    stub += "=" * len(name) + "\n\n"
    stub += STUB_TEMPLATE.format(module_name=SQL_MODULE_NAME, name=name)
    return stub


def generate_files(app):
    # Determine where to write .rst files to
    output_dir = os.path.join(app.srcdir, TOCTREE)
    os.makedirs(output_dir, exist_ok=True)

    # Write stubfiles
    func_names = get_sql_func_names()
    for name in func_names:
        stub_content = generate_stub(name)
        filename = f"{SQL_MODULE_NAME}.{name}.rst"
        filepath = os.path.join(output_dir, filename)
        with open(filepath, "w") as f:
            f.write(stub_content)

        # HACK: Not sure if this is ok?
        app.env.found_docs.add(filepath)


def setup(app):
    app.add_directive("sql-autosummary", SQLAutosummary)

    # Generate and register files when the builder is initialized
    app.connect("builder-inited", generate_files)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import pytest

import daft
from daft.sql import SQLCatalog


def test_joins_using():
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    df2 = daft.from_pydict({"idx": [1, 2], "score": [0.1, 0.2]})

    df_sql = daft.sql("select * from df1 join df2 using (idx)")
    actual = df_sql.collect().to_pydict()

    expected = df1.join(df2, on="idx").collect().to_pydict()

    assert actual == expected


def test_joins_with_alias():
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    df2 = daft.from_pydict({"idx": [1, 2], "score": [0.1, 0.2]})

    catalog = SQLCatalog({"df1": df1, "df2": df2})

    df_sql = daft.sql("select * from df1 as foo join df2 as bar on foo.idx=bar.idx where bar.score>0.1", catalog)

    actual = df_sql.collect().to_pydict()

    expected = {"idx": [2], "val": [20], "bar.idx": [2], "score": [0.2]}

    assert actual == expected


def test_joins_with_spaceship():
    df1 = daft.from_pydict({"idx": [1, 2, None], "val": [10, 20, 30]})
    df2 = daft.from_pydict({"idx": [1, 2, None], "score": [0.1, 0.2, None]})

    catalog = SQLCatalog({"df1": df1, "df2": df2})
    df_sql = daft.sql("select idx, val, score from df1 join df2 on (df1.idx<=>df2.idx)", catalog=catalog)

    actual = df_sql.collect().to_pydict()

    expected = {"idx": [1, 2, None], "val": [10, 20, 30], "score": [0.1, 0.2, None]}

    assert actual == expected


def test_joins_with_wildcard_expansion():
    df1 = daft.from_pydict({"idx": [1, 2], "val": [10, 20]})
    df2 = daft.from_pydict({"idx": [3], "score": [0.1]})
    df3 = daft.from_pydict({"idx": [1], "score": [0.1], "a": [1], "b": [2], "c": [3]})

    catalog = SQLCatalog({"df1": df1, "df2": df2, "df3": df3})

    df_sql = (
        daft.sql(
            """
        select df3.*
        from df1
        left join df2 on (df1.idx=df2.idx)
        left join df3 on (df1.idx=df3.idx)
        """,
            catalog,
        )
        .collect()
        .to_pydict()
    )

    expected = {"idx": [1, None], "score": [0.1, None], "a": [1, None], "b": [2, None], "c": [3, None]}

    assert df_sql == expected
    # make sure it works with exclusion patterns too

    df_sql = (
        daft.sql("""
        select df3.* EXCLUDE (a,b,c)
        from df1
        left join df2 on (df1.idx=df2.idx)
        left join df3 on (df1.idx=df3.idx)
        """)
        .collect()
        .to_pydict()
    )

    expected = {"idx": [1, None], "score": [0.1, None]}

    assert df_sql == expected


def test_joins_with_duplicate_columns():
    table1 = daft.from_pydict({"id": [1, 2, 3, 4], "value": ["a", "b", "c", "d"]})

    table2 = daft.from_pydict({"id": [2, 3, 4, 5], "value": ["b", "c", "d", "e"]})

    catalog = SQLCatalog({"table1": table1, "table2": table2})

    actual = daft.sql(
        """
        SELECT *
        FROM table1 t1
        LEFT JOIN table2 t2 on t2.id = t1.id
        ORDER BY t1.id;
        """,
        catalog,
    ).collect()

    expected = {
        "id": [1, 2, 3, 4],
        "value": ["a", "b", "c", "d"],
        "t2.id": [None, 2, 3, 4],
        "t2.value": [None, "b", "c", "d"],
    }

    assert actual.to_pydict() == expected


@pytest.mark.parametrize(
    "join_condition",
    [
        "x = y",
        "x = b.y",
        "y = x",
        "y = a.x",
        "a.x = y",
        "a.x = b.y",
        "b.y = x",
        "b.y = a.x",
    ],
)
@pytest.mark.parametrize("selection", ["*", "a.*, b.y, b.score", "a.x, a.val, b.*", "a.x, a.val, b.y, b.score"])
def test_join_qualifiers(join_condition, selection):
    a = daft.from_pydict({"x": [1, None], "val": [10, 20]})
    b = daft.from_pydict({"y": [1, None], "score": [0.1, 0.2]})

    catalog = SQLCatalog({"a": a, "b": b})

    df_sql = daft.sql(f"select {selection} from a join b on {join_condition}", catalog).to_pydict()

    expected = {"x": [1], "val": [10], "y": [1], "score": [0.1]}

    assert df_sql == expected


@pytest.mark.parametrize(
    "join_condition",
    [
        "x = y",
        "x = b1.y",
        "y = x",
        "y = a1.x",
        "a1.x = y",
        "a1.x = b1.y",
        "b1.y = x",
        "b1.y = a1.x",
    ],
)
@pytest.mark.parametrize(
    "selection", ["*", "a1.*, b1.y, b1.score", "a1.x, a1.val, b1.*", "a1.x, a1.val, b1.y, b1.score"]
)
def test_join_qualifiers_with_alias(join_condition, selection):
    a = daft.from_pydict({"x": [1, None], "val": [10, 20]})
    b = daft.from_pydict({"y": [1, None], "score": [0.1, 0.2]})

    catalog = SQLCatalog({"a": a, "b": b})

    df_sql = daft.sql(f"select {selection} from a as a1 join b as b1 on {join_condition}", catalog).to_pydict()

    expected = {"x": [1], "val": [10], "y": [1], "score": [0.1]}

    assert df_sql == expected


def test_cross_join():
    x = daft.from_pydict(
        {
            "A": [1, 3, 5],
            "B": ["a", "b", "c"],
        },
    )

    y = daft.from_pydict(
        {
            "C": [2, 4, 6, 8],
            "D": ["d", "e", "f", "g"],
        },
    )

    catalog = SQLCatalog({"x": x, "y": y})

    df = daft.sql("select * from x, y order by A, C", catalog)

    assert df.to_pydict() == {
        "A": [1, 1, 1, 1, 3, 3, 3, 3, 5, 5, 5, 5],
        "B": ["a", "a", "a", "a", "b", "b", "b", "b", "c", "c", "c", "c"],
        "C": [2, 4, 6, 8, 2, 4, 6, 8, 2, 4, 6, 8],
        "D": ["d", "e", "f", "g", "d", "e", "f", "g", "d", "e", "f", "g"],
    }



from __future__ import annotations

import datetime

import pyarrow as pa
import pytest

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_map_get():
    data = pa.array([[(1, 2)], [], [(2, 1)]], type=pa.map_(pa.int64(), pa.int64()))
    table = MicroPartition.from_arrow(pa.table({"map_col": data}))

    result = table.eval_expression_list([col("map_col").map.get(1)])

    assert result.to_pydict() == {"value": [2, None, None]}


def test_map_get_broadcasted():
    data = pa.array([[(1, 2)], [], [(2, 1)]], type=pa.map_(pa.int64(), pa.int64()))
    keys = pa.array([1, 3, 2], type=pa.int64())
    table = MicroPartition.from_arrow(pa.table({"map_col": data, "key": keys}))

    result = table.eval_expression_list([col("map_col").map.get(col("key"))])

    assert result.to_pydict() == {"value": [2, None, 1]}


def test_map_get_duplicate_keys():
    # Only the first value is returned
    data = pa.array([[(1, 2), (1, 3)]], type=pa.map_(pa.int64(), pa.int64()))
    table = MicroPartition.from_arrow(pa.table({"map_col": data}))

    result = table.eval_expression_list([col("map_col").map.get(1)])

    assert result.to_pydict() == {"value": [2]}


def test_map_get_logical_type():
    data = pa.array(
        [
            [("foo", datetime.date(2022, 1, 1))],
            [("foo", datetime.date(2022, 1, 2))],
            [],
        ],
        type=pa.map_(pa.string(), pa.date32()),
    )
    table = MicroPartition.from_arrow(pa.table({"map_col": data}))

    map = col("map_col").map
    result = table.eval_expression_list([map.get("foo")])

    assert result.to_pydict() == {"value": [datetime.date(2022, 1, 1), datetime.date(2022, 1, 2), None]}


def test_map_get_bad_field():
    data = pa.array([[(1, 2)], [(2, 3)]], type=pa.map_(pa.int64(), pa.int64()))
    table = MicroPartition.from_arrow(pa.table({"map_col": data}))

    with pytest.raises(ValueError):
        table.eval_expression_list([col("map_col").map.get("foo")])



# This file was copied from the Polars project (https://github.com/pola-rs/polars/blob/main/py-polars/debug/launch.py)
# under the license provided by Ritchie Vink and NVIDIA Corporation & Affiliates. The following parameter determines the sleep time of the Python process after a signal
# is sent that attaches the Rust LLDB debugger. If the Rust LLDB debugger attaches to the
# current session too late, it might miss any set breakpoints. If this happens
# consistently, it is recommended to increase this value.


import os
import re
import sys
import time
from pathlib import Path

LLDB_DEBUG_WAIT_TIME_SECONDS = 1


def launch_debugging() -> None:
    """Debug Rust files via Python.

    Determine the pID for the current debugging session, attach the Rust LLDB launcher,
    and execute the originally-requested script.
    """
    if len(sys.argv) == 1:
        msg = (
            "launch.py is not meant to be executed directly; please use the `Python: "
            "Debug Rust` debugging configuration to run a python script that uses the "
            "polars library."
        )
        raise RuntimeError(msg)

    # Get the current process ID.
    pID = os.getpid()

    # Print to the debug console to allow VSCode to pick up on the signal and start the
    # Rust LLDB configuration automatically.
    launch_file = Path(__file__).parents[1] / ".vscode/launch.json"
    if not launch_file.exists():
        msg = f"Cannot locate {launch_file}"
        raise RuntimeError(msg)
    with launch_file.open("r") as f:
        launch_info = f.read()

    # Overwrite the pid found in launch.json with the pid for the current process.
    # Match the initial "Rust LLDB" definition with the pid defined immediately after.
    pattern = re.compile('("Rust LLDB",\\s*"pid":\\s*")\\d+(")')
    found = pattern.search(launch_info)
    if not found:
        msg = (
            "Cannot locate pid definition in launch.json for Rust LLDB configuration. "
            "Please follow the instructions in the debugging section of the "
            "contributing guide (https://docs.pola.rs/development/contributing/ide/#debugging) "
            "for creating the launch configuration."
        )
        raise RuntimeError(msg)

    launch_info_with_new_pid = pattern.sub(rf"\g<1>{pID}\g<2>", launch_info)
    with launch_file.open("w") as f:
        f.write(launch_info_with_new_pid)

    # Print pID to the debug console. This auto-triggers the Rust LLDB configurations.
    print(f"pID = {pID}")

    # Give the LLDB time to connect. Depending on how long it takes for your LLDB
    # debugging session to initialize, you may have to adjust this setting.
    time.sleep(LLDB_DEBUG_WAIT_TIME_SECONDS)

    # Update sys.argv so that when exec() is called, the first argument is the script
    # name itself, and the remaining are the input arguments.
    sys.argv.pop(0)
    with Path(sys.argv[0]).open() as fh:
        script_contents = fh.read()

    # Run the originally requested file by reading in the script, compiling, and
    # executing the code.
    file_to_execute = Path(sys.argv[0])
    exec(compile(script_contents, file_to_execute, mode="exec"), {"__name__": "__main__"})


if __name__ == "__main__":
    launch_debugging()



from unittest.mock import MagicMock

import pytest

from daft.execution.physical_plan import Materialize, PartitionTask
from daft.execution.rust_physical_plan_shim import scan_with_tasks


def test_single_non_buffered_plan():
    result1, result2, result3 = MagicMock(), MagicMock(), MagicMock()
    scan = scan_with_tasks([MagicMock() for i in range(3)])
    materialized = Materialize(scan, results_buffer_size=None)
    plan = iter(materialized)

    # Buffer is unbounded and grows as we iterate on the plan (when no tasks are completed)
    assert len(materialized.materializations) == 0
    task1 = next(plan)
    assert isinstance(task1, PartitionTask)
    assert len(materialized.materializations) == 1
    task2 = next(plan)
    assert isinstance(task2, PartitionTask)
    assert len(materialized.materializations) == 2
    task3 = next(plan)
    assert isinstance(task3, PartitionTask)
    assert len(materialized.materializations) == 3

    # Ran out of work, waiting on new tasks to be completed
    assert next(plan) is None

    # Manually "complete" the tasks
    task1.set_result([result1])
    task1.set_done()
    task2.set_result([result2])
    task2.set_done()
    task3.set_result([result3])
    task3.set_done()

    # Results should be as we expect
    assert next(plan) == result1
    assert next(plan) == result2
    assert next(plan) == result3
    with pytest.raises(StopIteration):
        next(plan)


def test_single_non_buffered_plan_done_while_planning():
    result1, result2, result3 = MagicMock(), MagicMock(), MagicMock()
    scan = scan_with_tasks([MagicMock() for i in range(3)])
    materialized = Materialize(scan, results_buffer_size=None)
    plan = iter(materialized)

    # Buffer is unbounded and grows as we iterate on the plan (when no tasks are completed)
    assert len(materialized.materializations) == 0
    task1 = next(plan)
    assert isinstance(task1, PartitionTask)
    assert len(materialized.materializations) == 1
    task2 = next(plan)
    assert isinstance(task2, PartitionTask)
    assert len(materialized.materializations) == 2

    # Manually "complete" the tasks
    task1.set_result([result1])
    task1.set_done()
    task2.set_result([result2])
    task2.set_done()

    # On the next iteration, we should receive the result
    assert next(plan) == result1
    assert next(plan) == result2

    # Add more work to completion
    task3 = next(plan)
    assert isinstance(task3, PartitionTask)
    assert len(materialized.materializations) == 1  # only task3 on the buffer

    # Manually "complete" the last task
    task3.set_result([result3])
    task3.set_done()

    # Results should be as we expect
    assert next(plan) == result3
    with pytest.raises(StopIteration):
        next(plan)


def test_single_plan_with_buffer_slow_tasks():
    result1, result2, result3 = MagicMock(), MagicMock(), MagicMock()
    scan = scan_with_tasks([MagicMock() for i in range(3)])
    materialized = Materialize(scan, results_buffer_size=2)
    plan = iter(materialized)

    # Buffer and grows to size 2 (tasks are "slow" and don't complete faster than the next plan loop call)
    assert len(materialized.materializations) == 0
    task1 = next(plan)
    assert isinstance(task1, PartitionTask)
    assert len(materialized.materializations) == 1
    task2 = next(plan)
    assert isinstance(task2, PartitionTask)
    assert len(materialized.materializations) == 2

    # Plan cannot make forward progress until task1 finishes
    assert next(plan) is None
    task2.set_result([result2])
    task2.set_done()
    assert next(plan) is None
    task1.set_result([result1])
    task1.set_done()

    # Plan should fill its buffer with new tasks before starting to yield results again
    task3 = next(plan)
    assert isinstance(task3, PartitionTask)
    assert next(plan) == result1
    assert next(plan) == result2

    # Finish the last task
    task3.set_result([result3])
    task3.set_done()
    assert next(plan) == result3

    with pytest.raises(StopIteration):
        next(plan)


def test_single_plan_with_buffer_saturation_fast_tasks():
    result1, result2, result3 = MagicMock(), MagicMock(), MagicMock()
    scan = scan_with_tasks([MagicMock() for i in range(3)])
    materialized = Materialize(scan, results_buffer_size=2)
    plan = iter(materialized)

    # Buffer grows to size 1
    assert len(materialized.materializations) == 0
    task1 = next(plan)
    assert isinstance(task1, PartitionTask)
    assert len(materialized.materializations) == 1

    # Finish up on task 1 (task is "fast" and completes so quickly even before the next plan loop call)
    task1.set_result([result1])
    task1.set_done()

    # Plan should fill its buffer completely with new tasks before starting to yield results again
    task2 = next(plan)
    assert isinstance(task2, PartitionTask)
    assert len(materialized.materializations) == 1  # task1 has been popped
    task3 = next(plan)
    assert isinstance(task3, PartitionTask)
    assert len(materialized.materializations) == 2  # task1 has been popped
    assert next(plan) == result1
    assert next(plan) is None

    # Finish the last task(s)
    task2.set_result([result2])
    task2.set_done()
    task3.set_result([result3])
    task3.set_done()
    assert next(plan) == result2
    assert next(plan) == result3

    with pytest.raises(StopIteration):
        next(plan)
